package ri.stdf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Scanner;

import ri.util.ByteBuffer;
import ri.util.DialogSqliteConnect;
import ri.util.Emit;

/**
 * This class examines a supplied directory and looks for files
 * ending in .ritdb.  For each found file it creates two files
 * an metadata file with format and name the same as guru attribute files
 * and a body file whose name is its sha1 hash.
 * the original file is renamed to be the body file or its deleted
 * This class needs a guruID and a cert to sign the attributes with
 * @author markroos
 *
 */
public class RitdbMetadataCreator {
  File _directory;
  
  public RitdbMetadataCreator(File directory) {
    _directory = directory;
    createMetadata();
  }

  /**  Creates guru attributes and object files from RITDB files
  */

  public void createMetadata() {
    RitdbIdentityGenerator.getGuruId();  // read the key file
    Emit.out("Creating metadata files");
    File[] list=_directory.listFiles(new FilenameFilter() {
      public boolean accept(File dir, String name) {
        return name.toLowerCase().endsWith(".ritdb");
      }
    });
    int len=(list == null) ? 0 : list.length;
    for (int i=0; i < len; i++) {
      Emit.outNoCr(".");
      processFile(list[i]);
    }
    Emit.out("");
    Emit.out("Finished creating metadata");
    
  }

  private void processFile(File file) {
    
    // Generate the SHA object hash:
    byte[] hashbytes=RitdbIdentityGenerator.guruHash(MetadataIndexer.getFile(file));
    String objHash = RitdbIdentityGenerator.hexIt(hashbytes); //hash encoded as hex values
    File f = new File(_directory.getAbsolutePath() + "/" + objHash);
    FileOutputStream dout=null;
    try {
      dout=new FileOutputStream(f);
      dout.write(MetadataIndexer.getFile(file));
    }
    catch (IOException e) {
      Emit.out("Error in "+getClass().getName()+".processFile(): Unable to create body bfile "+f.getAbsolutePath()+" because "+e);
    }
    finally {
      try { dout.close(); } catch (Exception e) {}
    }

    //query the database file to create a list of key values

    HashMap<String, String> keyVals=new HashMap<String, String>(32, 0.9f); // attributes hash map
    String cid = RitdbIdentityGenerator.nextCid();
    keyVals.put("ri.sys.Cid" , cid );
    keyVals.put("ri.sys.RevisionOf",  cid);
    keyVals.put("ri.sys.Obsoleted", "false");
    keyVals.put("ri.sys.ObjClass", "RitdbLog");
    keyVals.put("ri.sys.ObjFormat", "ritdb");
    keyVals.put("ri.sys.ObjFormatVer", "1.0");
    keyVals.put("ri.sys.ObjHash", objHash);

    Connection sqlConnection=null;
    Statement sqlStatement=null;
    ResultSet results=null;
    HashMap<String,String> metadata = readMetadataDefFile();
    
    try {
      sqlConnection=DialogSqliteConnect.doConnectTo(file.getAbsolutePath(), false);
      sqlStatement=sqlConnection.createStatement();

      // get all of the run information
      results=sqlStatement.executeQuery("select name, value from narrow where entityID=1;");
      while (results.next()) {
        String attrName=results.getString(1);
        Object attrValue=results.getObject(2);
        if(metadata.containsKey(attrName)){
            keyVals.put(metadata.get(attrName) ,  attrValue.toString());
          }
        if (attrName.equals("STDF_FILE")) {
          keyVals.put("ri.sys.Title" , (String) attrValue);
        }
        if (attrName.equals("RUN_START_DTTM")) {
          keyVals.put("ri.sys.CreationDate" , (String) attrValue);
        }
      }
      results.close();
      results=sqlStatement.executeQuery("select min(value) from narrow where name='PART_ID';");
      results.next();
      int max = results.getInt(1);
      keyVals.put("dlog.StartDevice" , String.valueOf(max));
      results.close();
      results=sqlStatement.executeQuery("select max(value) from narrow where name='PART_ID';");
      results.next();
      max = results.getInt(1);
      keyVals.put("dlog.EndDevice" , String.valueOf(max));
    }
    catch (Exception e) {
      e.printStackTrace();
      Emit.out("Error in "+getClass().getName()+".processFile(): Unable to create metadata " + file.getAbsolutePath() + " because: " + e);
    }
    finally {
      try { if (results != null) results.close(); } catch (Exception e) {}
      try { if (sqlStatement != null) sqlStatement.close(); } catch (Exception e) {}
      try { if (sqlConnection != null) sqlConnection.close(); } catch (Exception e) {}
    }
    //  now the identity keys
    HashMap<String, String> identity=new HashMap<String, String>(32, 0.9f); // identity hash map
    identity.put("Cid", keyVals.get("ri.sys.Cid"));
    identity.put("Obsoleted", "false");   
    identity.put("ObjClass", keyVals.get("ri.sys.ObjClass"));
    identity.put("Title", keyVals.get("ri.sys.Title"));
    identity.put("RevisionOf", keyVals.get("ri.sys.RevisionOf"));
    identity.put("Sequence", ((String)keyVals.get("ri.sys.Cid")).substring(8, 16));
    identity.put("Owner", ((String)keyVals.get("ri.sys.Cid")).substring(0, 8));
    identity.put("CreationDate", ((String)keyVals.get("ri.sys.CreationDate")).substring(0, 8));
    identity.put("ObjHash", (String)keyVals.get("ri.sys.ObjHash"));

    

    // now we create a byte array of the identity and attributes
    ByteBuffer buffer=new ByteBuffer();
    buffer.writeTag(129);
    buffer.writeHashMap(identity);
    buffer.writeTag(130);
    buffer.writeHashMap(keyVals);
    //add security keys, adds signer and sign, attrHash
    RitdbIdentityGenerator.addSecurityAttributes(buffer, hashbytes);

    //now write out the .gat file
    //compute the hash of the gat bytes as a file name
    byte[] attrhashbytes=RitdbIdentityGenerator.guruHash(buffer.toBytes());
    String attrHash = RitdbIdentityGenerator.hexIt(attrhashbytes); //hash encoded as hex values
    f = new File(_directory.getAbsolutePath() + "/" + attrHash + ".gat");
    dout=null;
    try {
      dout=new FileOutputStream(f);
      dout.write(buffer.toBytes());
    }
    catch (IOException e) {
      Emit.out("Error in "+getClass().getName()+".processFile(): Unable to create file "+f.getAbsolutePath()+" because "+e);
    }
    finally {
      try { dout.close(); } catch (Exception e) {}
    }
    file.renameTo(new File(file.getAbsolutePath().substring(0, file.getAbsolutePath().lastIndexOf('.')+1)+"sqlite"));
    //System.out.println("verify gat "+ RitdbIdentityGenerator.verifyGat(bb.toBytes()));
  }
   
  private HashMap<String,String> readMetadataDefFile(){
    HashMap<String,String> metadata = new HashMap<String,String>() ;
    File file = new File("metadata.txt");
    if (!file.exists()) return null; //no such file found

    Scanner din = null;
    try {
      din=new Scanner(file);
      while (din.hasNext()){
        String[] next = din.nextLine().split(",");
        metadata.put(next[1], next[0]);
      }
    }
    catch (FileNotFoundException e) {
    }
    finally{
      din.close();
    }
    return metadata;
    }
  
}
