package ri.stdf;

import java.io.File;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;
import java.util.TimeZone;

import ri.util.DialogSqliteConnect;
import ri.util.Emit;

import com.tragicphantom.stdf.Record;
import com.tragicphantom.stdf.RecordData;
import com.tragicphantom.stdf.RecordDescriptor;
import com.tragicphantom.stdf.RecordType;
import com.tragicphantom.stdf.STDFWriter;
import com.tragicphantom.stdf.v4.Types;

/**Saves to an STDF4 file, using a separate thread
 * This saves the contents of a narrow database table to STDF4 file format on disk
 * */
public class SaveStdf4Thread extends Thread {
  private Connection _sqlConnection;  //connection to database
  private File _saveFile;             //file to save to
  private StdfUpload _parent;         //parent window, for settings and display purposes
  private static Map<String, RecordDescriptor> stdfTypes =   //map of STDF4 types 
      createTypeMap(Types.getRecordDescriptors());
  private ArrayList<Record> records=new ArrayList<Record>();
  private Statement sqlStatement = null;
  private Object[] sdrValues = null;
  RecordDescriptor sdrDesc= null;
  RecordDescriptor mrrDesc= null;
  private Object[] mrrValues = null;
  private Object[] wrrValues = null;
  Long lotEndTime ;
  private Hashtable<Integer,Integer> testNumbers = new Hashtable<Integer,Integer>();
  
  /**Constr*/
  public SaveStdf4Thread(StdfUpload parent, File sourceFile, File saveFile) {
    _parent = parent;
    try {
		_sqlConnection = DialogSqliteConnect.doConnectTo(sourceFile.getAbsolutePath(), false);
	} 	catch (ClassNotFoundException e) {
		e.printStackTrace();
	}
		catch ( SQLException e) {
		e.printStackTrace();
	}
    _saveFile=saveFile;
    records=new ArrayList<Record>();
    try {
    	Statement tmpStatement = _sqlConnection.createStatement();
    	tmpStatement.execute("CREATE INDEX IF NOT EXISTS main.tmpEids ON narrow (entityID ASC);");
    	tmpStatement.execute("CREATE INDEX IF NOT EXISTS main.tmpNames ON narrow (value ASC);");
    	tmpStatement.close();
	} catch (SQLException e) {
		e.printStackTrace();
	}

  }
  
  @Override
  public void run() {
    System.out.println("Saving STDF4 to "+_saveFile.getName());
    try {
      records.add(createRecord("FAR", 2, 4));
      // ---------------------------------
      // Example: accessing the database:
      // ---------------------------------
      sqlStatement = _sqlConnection.createStatement();
      
      //First get the MIR information
      generateMir();
      generateWcr();
      generateSdr();
      generateWir();
      generatePtr();
      generateTsr();
      generateHbr();
      generateSbr();
      generatePcr();
      generateWrr();
      generateMrr(); 
  
      // ... the other records
      
      STDFWriter writer=new STDFWriter(_saveFile);
      writer.write(records);
      writer.close();
	  sqlStatement.execute("drop index if exists tmpNames;");
	  sqlStatement.execute("drop index if exists tmpEids;");
	  sqlStatement.execute("vacuum;");
      sqlStatement.close();
      _sqlConnection.close();
      
      System.out.println("All done.");
    }
    catch(Exception e) {
      String msg = "Error saving to STDF4 file "+_saveFile.getName()+": "+e.getMessage(); 
      Emit.out(msg);
      System.out.println(msg);
      e.printStackTrace();
    }
  }
  
  	private  long convertTimeToUnix(String time){
      DateFormat dfm = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");  

      long unixtime = 0;
      dfm.setTimeZone(TimeZone.getTimeZone("GMT"));//Specify your timezone 
      try
      {
          unixtime = dfm.parse(time).getTime();  
          unixtime=unixtime/1000;
      } 
      catch (Exception e) 
      {
          e.printStackTrace();
      }
      return unixtime;
      }

  
  /**Utility method to create the STDF4 types map*/
  private static Map<String, RecordDescriptor> createTypeMap(Map<RecordType, RecordDescriptor> spec) {
    Map<String, RecordDescriptor> types=new HashMap<String, RecordDescriptor>();
    for(RecordDescriptor desc : spec.values())
      types.put(desc.getType().toUpperCase(), desc);
    return types;
  }

  /**Utility method to create an STDF4 record*/
  private Record createRecord(String type, Object... values) {
    RecordDescriptor desc=stdfTypes.get(type);
    //Assert.assertEquals(desc.size(), values.length);
    Object[] fields=new Object[desc.size()];
    for(int i=0; i < values.length && i < desc.size(); i++)
      fields[i]=values[i];
    return new Record(desc, new RecordData(desc, fields));
  }
  /**Utility method to create an default STDF4 record*/
  private Object[] createDefaultRecord(RecordDescriptor desc, Object... values) {
    Object[] fields=new Object[desc.size()];
    for(int i=0; i < values.length && i < desc.size(); i++)
      fields[i]=values[i];
    return fields;
  }

  /**Returns current time (in seconds)*/
  private static long today() {
    return Calendar.getInstance().getTimeInMillis() / 1000;
  }
  
  private void generateMir() throws SQLException{
		// need a holder for SDR info that is put into the MIR
	sdrValues = new Object[(stdfTypes.get("SDR")).size()];
	sdrDesc=stdfTypes.get("SDR");
	mrrDesc=stdfTypes.get("MRR");
	mrrValues = new Object[(stdfTypes.get("MRR")).size()];
    //First get the MIR information
    RecordDescriptor desc=stdfTypes.get("MIR");
    Object[] fieldValues = createDefaultRecord(desc, 
        0, 0, 1, " ", " ", " ", 65535, 
        " ", "lotid", "", "", "", "", "", "", "", "", "", "", "", "", "", "", 
        "", "", "", "", "", "", "", "", "", "", "", "", "", "", "");
    // get the MIR items entityID = 1 and iterate over them
    ResultSet results = sqlStatement.executeQuery("select name, value from narrow where entityID=1;");
    while(results.next()){
      String attrName = results.getString(1);
      Object attrValue = results.getObject(2);
      //if( attrName.equals("SETUP_T"))fieldValues[desc.getIndex("SETUP_T")] =  attrValue; 

      if( attrName.equals("TESTER_ID"))fieldValues[desc.getIndex("NODE_NAM")] = attrValue;   
      if( attrName.equals("LOT_ID"))fieldValues[desc.getIndex("LOT_ID")] = attrValue;
      if( attrName.equals("STATION_NUMBER"))fieldValues[desc.getIndex("STAT_NUM")] = attrValue;
      if( attrName.equals("TEST_MODE_CODE"))fieldValues[desc.getIndex("MODE_COD")] = attrValue;
      if( attrName.equals("RETEST_CODE"))fieldValues[desc.getIndex("RTST_COD")] = attrValue;
      if( attrName.equals("DATA_PROTECTION_CODE"))fieldValues[desc.getIndex("PROT_COD")] = attrValue;
      if( attrName.equals("COMMAND_MODE_CODE"))fieldValues[desc.getIndex("CMOD_COD")] = attrValue;   
      if( attrName.equals("TESTER_TYPE"))fieldValues[desc.getIndex("TSTR_TYP")] = attrValue;
      if( attrName.equals("PRODUCT_ID"))fieldValues[desc.getIndex("PART_TYP")] = attrValue;
      if( attrName.equals("JOB_NAME"))fieldValues[desc.getIndex("JOB_NAM")] = attrValue;    
      if( attrName.equals("JOB_REVISION"))fieldValues[desc.getIndex("JOB_REV")] = attrValue.toString();
      if( attrName.equals("STEP_NAME"))fieldValues[desc.getIndex("SBLOT_ID")] = attrValue;
      if( attrName.equals("OPERATOR_ID"))fieldValues[desc.getIndex("OPER_NAM")] = attrValue;    
      if( attrName.equals("TESTER_EXEC_SW_TYPE"))fieldValues[desc.getIndex("EXEC_TYP")] = attrValue;
      if( attrName.equals("TESTER_EXEC_SW_VER"))fieldValues[desc.getIndex("EXEC_VER")] = attrValue;
      if( attrName.equals("TEST_PASS_NAME"))fieldValues[desc.getIndex("TEST_COD")] = attrValue;
      if( attrName.equals("TEST_TEMPERATURE"))fieldValues[desc.getIndex("TST_TEMP")] = attrValue;
      if( attrName.equals("USER_TXT"))fieldValues[desc.getIndex("USER_TXT")] = attrValue;
      if( attrName.equals("AUXILIARY_FILE"))fieldValues[desc.getIndex("AUX_FILE")] = attrValue;    
      if( attrName.equals("PACKAGE_TYPE"))fieldValues[desc.getIndex("PKG_TYP")] = attrValue;
      if( attrName.equals("PRODUCT_FAMLY_ID"))fieldValues[desc.getIndex("FAMLY_ID")] = attrValue;
      if( attrName.equals("DATE_CODE"))fieldValues[desc.getIndex("DATE_COD")] = attrValue;
      if( attrName.equals("TEST_FACILITY_ID"))fieldValues[desc.getIndex("FACIL_ID")] = attrValue;
      if( attrName.equals("TEST_FLOOR_ID"))fieldValues[desc.getIndex("FLOOR_ID")] = attrValue;  
      if( attrName.equals("FAB_PROCESS_ID"))fieldValues[desc.getIndex("PROC_ID")] = attrValue;
      //if( attrName.equals("OPER_FRQ"))fieldValues[desc.getIndex("OPER_FRQ")] = attrValue;
      if( attrName.equals("TEST_SPECIFICATION_NAME"))fieldValues[desc.getIndex("SPEC_NAM")] = attrValue;
      if( attrName.equals("TEST_SPECIFICATION_VERSION"))fieldValues[desc.getIndex("SPEC_VER")] = attrValue;
      if( attrName.equals("FLOW_ID"))fieldValues[desc.getIndex("FLOW_ID")] = attrValue;
      if( attrName.equals("SETUP_ID"))fieldValues[desc.getIndex("SETUP_ID")] = attrValue;
      if( attrName.equals("DESIGN_REV"))fieldValues[desc.getIndex("DSGN_REV")] = attrValue;
      if( attrName.equals("ENGINEERING_LOT_ID"))fieldValues[desc.getIndex("ENG_ID")] = attrValue;
      if( attrName.equals("ROM_CODE"))fieldValues[desc.getIndex("ROM_COD")] = attrValue;
      if( attrName.equals("TESTER_SERIAL_NUMBER"))fieldValues[desc.getIndex("SERL_NUM")] = attrValue;
      if( attrName.equals("SUPERVISOR_ID"))fieldValues[desc.getIndex("SUPR_NAM")] = attrValue;
      if( attrName.equals("USER_JOB_DESCRIPTION"))fieldValues[desc.getIndex("USER_TXT")] = attrValue;
      if( attrName.equals("LOT_DISPOSITION_CODE"))fieldValues[desc.getIndex("DISP_COD")] = attrValue;
      if( attrName.equals("EXEC_LOT_DESCRIPTION"))fieldValues[desc.getIndex("EXC_DESC")] = attrValue;
      if( attrName.equals("RUN_START_DTTM"))fieldValues[desc.getIndex("START_T")] = (Long) convertTimeToUnix((String)attrValue);
      if( attrName.equals("JOB_SETUP_DTTM"))fieldValues[desc.getIndex("SETUP_T")] = (Long) convertTimeToUnix((String)attrValue);

      // values from sdr that are in run info
      if( attrName.equals("HANDLER_TYPE"))sdrValues[sdrDesc.getIndex("HAND_TYP")] = attrValue;
      if( attrName.equals("HANDLER_ID"))sdrValues[sdrDesc.getIndex("HAND_ID")] = attrValue;
      if( attrName.equals("CARD_TYPE"))sdrValues[sdrDesc.getIndex("CARD_TYP")] = attrValue;
      if( attrName.equals("CARD_ID"))sdrValues[sdrDesc.getIndex("CARD_ID")] = attrValue;
      if( attrName.equals("LOADBOARD_TYPE"))sdrValues[sdrDesc.getIndex("LOAD_TYP")] = attrValue;
      if( attrName.equals("LOADBOARD_ID"))sdrValues[sdrDesc.getIndex("LOAD_ID")] = attrValue;
      if( attrName.equals("DIB_TYPE"))sdrValues[sdrDesc.getIndex("DIB_TYP")] = attrValue;
      if( attrName.equals("DIB_ID"))sdrValues[sdrDesc.getIndex("DIB_ID")] = attrValue;
      if( attrName.equals("INTERFACE_CABLE_TYPE"))sdrValues[sdrDesc.getIndex("CABL_TYP")] = attrValue;
      if( attrName.equals("INTERFACE_CABLE_ID"))sdrValues[sdrDesc.getIndex("CABL_ID")] = attrValue;
      if( attrName.equals("CONTACTOR_TYPE"))sdrValues[sdrDesc.getIndex("CONT_TYP")] = attrValue;
      if( attrName.equals("CONTACTOR_ID"))sdrValues[sdrDesc.getIndex("CONT_ID")] = attrValue;
      if( attrName.equals("LASER_TYPE"))sdrValues[sdrDesc.getIndex("LASR_TYP")] = attrValue;
      if( attrName.equals("LASER_ID"))sdrValues[sdrDesc.getIndex("LASR_ID")] = attrValue;
      
      //MRR information
      if( attrName.equals("RUN_END_DTTM"))mrrValues[desc.getIndex("SETUP_T")] = (Long) convertTimeToUnix((String)attrValue);
      if( attrName.equals("LOT_DISPOSITION_CODE"))mrrValues[mrrDesc.getIndex("DISP_COD")] = attrValue;
      if( attrName.equals("USER_LOT_DESCRIPTION"))mrrValues[mrrDesc.getIndex("USR_DESC")] = attrValue;
      if( attrName.equals("EXEC_LOT_DESCRIPTION"))mrrValues[mrrDesc.getIndex("EXC_DESC")] = attrValue;
    }
    results.close();
    RecordData data = new RecordData(desc, fieldValues);
    Record record = new Record(desc,data );
    records.add(record);  
  }
  
  private void generateSdr() throws SQLException{
	  sdrValues[sdrDesc.getIndex("HEAD_NUM")] = 1;
	    // get the set of SDRs and iterate over them
	  ResultSet sdrs = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='SITE_INFO';");
	  while(sdrs.next()){
		  	int siteCount = 0;
		  	//TODO need to support multiple sites
		  	Object[] sites = new Object[1];
		  	// for each sdr get the values
		  	
		  	Object[] fieldValues = new Object[sdrValues.length];
		  	System.arraycopy(sdrValues, 0, fieldValues, 0, sdrValues.length);
	        ResultSet results = sqlStatement.executeQuery("select name, value, indexID from narrow where entityID="+(sdrs.getString(1))+";");
	        while(results.next()){
	          String attrName = results.getString(1);
	          Object attrValue = results.getObject(2);   
	          if( attrName.equals("SITE_GROUP_NUMBER"))sdrValues[sdrDesc.getIndex("SITE_GRP")] = attrValue;
	          if( attrName.equals("ACTIVE_SITE")){
	        	  results.getInt(3);
	        	  sites[0] = new Long((Integer) attrValue);
	        	  siteCount++;
	          }
	        }
	        sdrValues[sdrDesc.getIndex("SITE_CNT")] = new Long(siteCount);
	        sdrValues[sdrDesc.getIndex("SITE_NUM")] = sites;
	        results.close();
	        RecordData data = new RecordData(sdrDesc, sdrValues);
	        Record record = new Record(sdrDesc,data );
	        records.add(record);
	    }
        sdrs.close();
	    }
  
  private void generateTsr() throws SQLException{
	    // get the set of TSRs and iterate over them
	  RecordDescriptor desc=stdfTypes.get("TSR");
	  ResultSet tsrs = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='TEST_SUMMARY';");
	  while(tsrs.next()){
		  	boolean site = false;
		  	int options = 255;
			Object[] fieldValues =  new Object[(desc).size()];
        	fieldValues[desc.getIndex("TEST_TIM")] =  0.0;
        	fieldValues[desc.getIndex("TEST_TYP")] =  'P';
        	fieldValues[desc.getIndex("ALRM_CNT")] =  0;
		  	// for each tsr get the values
        	ResultSet results = sqlStatement.executeQuery("select name, value from narrow where entityID="+(tsrs.getString(1))+";");
	        while(results.next()){
	          String attrName = results.getString(1);
	          Object attrValue = results.getObject(2); 
	          if( attrName.equals("SITE")){
	        	  site = true;
	        	  fieldValues[desc.getIndex("SITE_NUM")] = attrValue;
	          }
	          if( attrName.equals("TEST_TYP"))fieldValues[desc.getIndex("TEST_TYP")] = attrValue;
	          if( attrName.equals("TEST_ID"))fieldValues[desc.getIndex("TEST_NUM")] = attrValue;
	          if( attrName.equals("TEST_EXECUTION_COUNT"))fieldValues[desc.getIndex("EXEC_CNT")] = attrValue;
	          if( attrName.equals("TEST_FAIL_COUNT"))fieldValues[desc.getIndex("FAIL_CNT")] = attrValue;
	          if( attrName.equals("TEST_ALARM_COUNT"))fieldValues[desc.getIndex("ALRM_CNT")] = attrValue;
	          if( attrName.equals("TSR_TEST_NAME"))fieldValues[desc.getIndex("TEST_NAM")] = attrValue;
	          if( attrName.equals("TEST_SEQUENCER_NAME"))fieldValues[desc.getIndex("SEQ_NAM")] = attrValue;
	          if( attrName.equals("TEST_LABEL"))fieldValues[desc.getIndex("TEST_LBL")] = attrValue;
	          if( attrName.equals("AVERAGE_TEST_TIME")){
	        	  fieldValues[desc.getIndex("TEST_TIM")] = attrValue;
	        	  options = options & 0xfb;
	          }
	          if( attrName.equals("MINIMUM_TEST_VALUE")){
	          	  fieldValues[desc.getIndex("TEST_MIN")] = results.getDouble(2);
	        	  options = options & 0xfe;
	          }
	          if( attrName.equals("MAXIMUM_TEST_VALUE")){
		          fieldValues[desc.getIndex("TEST_MAX")] = results.getDouble(2);
	        	  options = options & 0xfd;
	          }
	          if( attrName.equals("SUM_TEST_VALUES")){
		          fieldValues[desc.getIndex("TST_SUMS")] = results.getDouble(2);
	        	  options = options & 0xef;
	          }
	          if( attrName.equals("SUMOFSQUARES_TEST_VALUES")){
		          fieldValues[desc.getIndex("TST_SQRS")] = results.getDouble(2);
	        	  options = options & 0xdf;
	          }
	        }
	        results.close();
        	fieldValues[desc.getIndex("OPT_FLAG")] =  options;
        	fieldValues[desc.getIndex("TEST_TYP")] = " ";
	        if(site){
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 1;
	        }else{
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 255;
	        }
	        RecordData data = new RecordData(desc, fieldValues);
	        Record record = new Record(desc,data );
	        records.add(record);
	  	}
        tsrs.close();
	    }
  
  private void generateHbr() throws SQLException{
	    // get the set of hbrs and iterate over them
	  RecordDescriptor desc=stdfTypes.get("HBR");
	  ResultSet hbrs = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='HARD_BIN';");
	  while(hbrs.next()){
		  	boolean site = false;
		  	// for each hbr  get the values
			Object[] fieldValues =  new Object[(desc).size()];
	        ResultSet results = sqlStatement.executeQuery("select name, value from narrow where entityID="+(hbrs.getString(1))+";");
	        while(results.next()){
	          String attrName = results.getString(1);
	          Object attrValue = results.getObject(2);   
	          if( attrName.equals("SITE")){
	        	  site = true;
	        	  fieldValues[desc.getIndex("SITE_NUM")] = attrValue;
	          }
	          if( attrName.equals("BIN_NUMBER"))fieldValues[desc.getIndex("HBIN_NUM")] = attrValue;
	          if( attrName.equals("BIN_COUNT"))fieldValues[desc.getIndex("HBIN_CNT")] = attrValue;
	          if( attrName.equals("BIN_PF"))fieldValues[desc.getIndex("HBIN_PF")] = attrValue;
	          if( attrName.equals("BIN_NAME"))fieldValues[desc.getIndex("HBIN_NAM")] = attrValue;

	        }
	        if(site){
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 1;
	        }else{
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 255;
	        }
	        results.close();
	        RecordData data = new RecordData(desc, fieldValues);
	        Record record = new Record(desc,data );
	        records.add(record);
	    }
      hbrs.close();
	    }
  
  private void generateSbr() throws SQLException{
	    // get the set of sbrs and iterate over them
	  RecordDescriptor desc=stdfTypes.get("SBR");
	  ResultSet sbrs = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='SOFT_BIN';");
	  while(sbrs.next()){
		  	// for each sbr  get the values
		  	boolean site = false;
			Object[] fieldValues =  new Object[(desc).size()];
	        ResultSet results = sqlStatement.executeQuery("select name, value from narrow where entityID="+(sbrs.getString(1))+";");
	        while(results.next()){
	          String attrName = results.getString(1);
	          Object attrValue = results.getObject(2);   
	          if( attrName.equals("SITE")){
	        	  site = true;
	        	  fieldValues[desc.getIndex("SITE_NUM")] = attrValue;
	          }
	          if( attrName.equals("BIN_NUMBER"))fieldValues[desc.getIndex("SBIN_NUM")] = attrValue;
	          if( attrName.equals("BIN_COUNT"))fieldValues[desc.getIndex("SBIN_CNT")] = attrValue;
	          if( attrName.equals("BIN_PF"))fieldValues[desc.getIndex("SBIN_PF")] = attrValue;
	          if( attrName.equals("BIN_NAME"))fieldValues[desc.getIndex("SBIN_NAM")] = attrValue;

	        }
	        if(site){
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 1;
	        }else{
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 255;
	        }
	        results.close();
	        RecordData data = new RecordData(desc, fieldValues);
	        Record record = new Record(desc,data );
	        records.add(record);
	    }
    sbrs.close();
	    }
  
  private void generatePcr() throws SQLException{
	    // get the set of pcrs and iterate over them
	  RecordDescriptor desc=stdfTypes.get("PCR");
	  ResultSet pcrs  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='RUN_SUMMARY';");
	  while(pcrs.next()){
		  	// for each sbr  get the values
		  	boolean site = false;
		  	Object[] fieldValues =  new Object[(desc).size()];
		  	fieldValues[desc.getIndex("RTST_CNT")] =  0;
	        ResultSet results = sqlStatement.executeQuery("select name, value from narrow where entityID="+(pcrs.getString(1))+";");
	        while(results.next()){
	          String attrName = results.getString(1);
	          Object attrValue = results.getObject(2);   
	          if( attrName.equals("PART_COUNT"))fieldValues[desc.getIndex("PART_CNT")] = attrValue;
	          if( attrName.equals("RETEST_PART_COUNT"))fieldValues[desc.getIndex("RTST_CNT")] = attrValue;
	          if( attrName.equals("ABORT_PART_COUNT"))fieldValues[desc.getIndex("ABRT_CNT")] = attrValue;
	          if( attrName.equals("GOOD_PART_COUNT")){
	        	  fieldValues[desc.getIndex("GOOD_CNT")] = attrValue;
	        	  fieldValues[desc.getIndex("FUNC_CNT")] = attrValue;
	          }
	          if( attrName.equals("SITE")){
	        	  site = true;
	        	  fieldValues[desc.getIndex("SITE_NUM")] = attrValue;
	          }
	        }
	        results.close();
	        if(site){
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 1;
	        }else{
	        	fieldValues[desc.getIndex("HEAD_NUM")] = 255;
	        }
	        RecordData data = new RecordData(desc, fieldValues);
	        Record record = new Record(desc,data );
	        records.add(record);
	    }
	  pcrs.close();
	    }
  
  private void generatePtr() throws SQLException{
	    // get the test events and create pir, ptr and prr as needed
	  RecordDescriptor pirDesc=stdfTypes.get("PIR");
	  RecordDescriptor prrDesc=stdfTypes.get("PRR");
	  RecordDescriptor ptrDesc=stdfTypes.get("PTR");
	  // get the eid for the limit file
	  ResultSet limitName  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='TEST_LIMITS';");
	  limitName.next();
	  String limitSet = limitName.getString(1);
	  limitName.close();
	  // now get the eids for all of the test events
	  boolean firstTime = true;
	  ResultSet events  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='TEST_EVENT';");
	  while(events.next()){
		  	// build the PIR and PRR fields
		  Object[] pirValues =  new Object[(pirDesc).size()];
		  Object[] prrValues =  new Object[(prrDesc).size()];
		  pirValues[pirDesc.getIndex("HEAD_NUM")] = 1;
		  prrValues[prrDesc.getIndex("HEAD_NUM")] = 1;
		  prrValues[prrDesc.getIndex("X_COORD")] = -32768;
		  prrValues[prrDesc.getIndex("Y_COORD")] = -32768;
		  prrValues[prrDesc.getIndex("PART_FIX")] = new byte[0];
		  Long site = 1L;
    	  int flags = 0;
	      ResultSet pir = _sqlConnection.createStatement().executeQuery("select name, value from narrow where indexID=0 and entityID="+(events.getString(1))+";");
	      while(pir.next()){
	          String attrName = pir.getString(1);
	          Object attrValue = pir.getObject(2); 
	          if( attrName.equals("SITE")){
	        	  site = pir.getLong(2);
	        	  pirValues[pirDesc.getIndex("SITE_NUM")] = site;
	        	  prrValues[prrDesc.getIndex("SITE_NUM")] = site;
	          }
	          if( attrName.equals("NUM_TESTS"))prrValues[prrDesc.getIndex("NUM_TEST")] = attrValue;
	          if( attrName.equals("HARD_BIN"))prrValues[prrDesc.getIndex("HARD_BIN")] = attrValue;
	          if( attrName.equals("SOFT_BIN"))prrValues[prrDesc.getIndex("SOFT_BIN")] = attrValue;
	          if( attrName.equals("X"))prrValues[prrDesc.getIndex("X_COORD")] = attrValue;
	          if( attrName.equals("Y"))prrValues[prrDesc.getIndex("Y_COORD")] = attrValue;
	          if( attrName.equals("UNIT_TEST_TIME")){
	        	  Long time = (long)((pir.getDouble(2)) * 1000.0);  // time in ms
	        	  prrValues[prrDesc.getIndex("TEST_T")] = time;
	          }
	          if( attrName.equals("PART_ID"))prrValues[prrDesc.getIndex("PART_ID")] = pir.getString(2);
	          if( attrName.equals("PART_TEXT"))prrValues[prrDesc.getIndex("PART_TXT")] = attrValue;
	          if( attrName.equals("RETEST_CODE")) flags = flags | 0x1  ;
	          if( attrName.equals("PF")){
	        	  if(((String) attrValue).equals("PASS")) flags = flags | 0x0;
	        	  if(((String) attrValue).equals("FAIL")) flags = flags | 0x8;
	        	  if(((String) attrValue).equals("TESTED")) flags = flags | 0x10;
	        	  if(((String) attrValue).equals("ABORT")) flags = flags | 0x4;
	           prrValues[prrDesc.getIndex("PART_FLG")] = flags;  
	          }
	      }
	      pir.close();
		  	// write the PIR
	      RecordData data = new RecordData(pirDesc, pirValues);
	      Record record = new Record(pirDesc,data );
	      records.add(record);
	      Object[] ptrValues =  null;
		  	// for each event  get the PTR values
	        ResultSet results = sqlStatement.executeQuery("select indexID, value, value2 from narrow where name='R' and entityID="+(events.getString(1))+";");
	        while(results.next()){
	      	  // if this is the first time get the full info for the testInfo
  	          if(firstTime){
  		      	  ptrValues =  new Object[(ptrDesc).size()];
  	 	          ptrValues[ptrDesc.getIndex("RES_SCAL")] = 0;
  	 	          ptrValues[ptrDesc.getIndex("LLM_SCAL")] = 0;
  	 	          ptrValues[ptrDesc.getIndex("HLM_SCAL")] = 0;
  	  	          ptrValues[ptrDesc.getIndex("HI_LIMIT")] = 0.0;
  	  	          ptrValues[ptrDesc.getIndex("LO_LIMIT")] = 0.0;
  	  	          ptrValues[ptrDesc.getIndex("LO_SPEC")] = 0.0;
  	  	          ptrValues[ptrDesc.getIndex("HI_SPEC")] = 0.0;
  	        	int optFlag = 255;
		        ResultSet testInfo = _sqlConnection.createStatement().executeQuery("select name, value  from narrow where entityID="+(results.getString(1))+";");
		        while(testInfo.next()){
		        	String attrName = testInfo.getString(1);
		        	Object attrValue = testInfo.getObject(2);
			        if( attrName.equals("TEST_NUMBER")){
			        	ptrValues[ptrDesc.getIndex("TEST_NUM")] = attrValue;
			        	testNumbers.put(results.getInt(1), testInfo.getInt(2));
			        }
			        if( attrName.equals("TEST_NAME"))ptrValues[ptrDesc.getIndex("TEST_TXT")] = attrValue;
			        if( attrName.equals("ALARM_ID"))ptrValues[ptrDesc.getIndex("ALARM_ID")] = attrValue;
			        if( attrName.equals("R_SCALE")){
			        	optFlag = optFlag & 0xfe;
			        	ptrValues[ptrDesc.getIndex("RES_SCAL")] = attrValue;
				        ptrValues[ptrDesc.getIndex("LLM_SCAL")] = attrValue;
				        ptrValues[ptrDesc.getIndex("HLM_SCAL")] = attrValue;
			        }
			        if( attrName.equals("TEST_UNITS"))ptrValues[ptrDesc.getIndex("UNITS")] = attrValue;
			        ptrValues[ptrDesc.getIndex("OPT_FLAG")] = optFlag;
		        }
		        testInfo.close();
		        // LIMITS
		        ResultSet limits = _sqlConnection.createStatement().executeQuery("select name, value from narrow where entityID="+limitSet+" and indexID="+(results.getString(1))+";");
		        while(limits.next()){
		        	String attrName = limits.getString(1);
			        if( attrName.equals("UL")){
			        	optFlag = optFlag & 0x5f;
			        	ptrValues[ptrDesc.getIndex("HI_LIMIT")] = limits.getDouble(2);
			        }
			        if( attrName.equals("LL")){
			        	optFlag = optFlag & 0xaf;
			        	ptrValues[ptrDesc.getIndex("LO_LIMIT")] = limits.getDouble(2);
			        }
			        if( attrName.equals("USL")){
			        	optFlag = optFlag & 0xf7;
			        	ptrValues[ptrDesc.getIndex("HI_SPEC")] = limits.getDouble(2);
			        }
			        if( attrName.equals("LSL")){
			        	optFlag = optFlag & 0xfb;
			        	ptrValues[ptrDesc.getIndex("LO_SPEC")] = limits.getDouble(2);
			        }
		        }
		        limits.close();
		        ptrValues[ptrDesc.getIndex("OPT_FLAG")] = optFlag;
  	          }else{
  	        	ptrValues =  new Object[8];  // don't write the optional data
		        ptrValues[ptrDesc.getIndex("TEST_NUM")] = testNumbers.get(results.getInt(1));
  	          }
		    // now insert the data and the flag
	        ptrValues[ptrDesc.getIndex("SITE_NUM")] = site;
	        ptrValues[ptrDesc.getIndex("HEAD_NUM")] = 1;
	  	    ptrValues[ptrDesc.getIndex("RESULT")] = 0.0;  
	        ptrValues[ptrDesc.getIndex("RESULT")] = results.getDouble(2);
	        flags = 0;
	        if(results.getString(3).equals("P")) flags = 0;
	        if(results.getString(3).equals("F")) flags = 128;
	        if(results.getString(3) == null) flags = 64;
	        ptrValues[ptrDesc.getIndex("TEST_FLG")] = flags;
	        ptrValues[ptrDesc.getIndex("PARM_FLG")] = 0;
	        data = new RecordData(ptrDesc, ptrValues);
	        record = new Record(ptrDesc,data );
	        records.add(record); 
	        }
		    results.close();       
		  	// write the PRR
        	firstTime = false;
	        data = new RecordData(prrDesc, prrValues);
	        record = new Record(prrDesc,data );
	        records.add(record);
	  }
	  events.close();
  }
  
  private void generateMrr() throws SQLException{
	    // get the set of TSRs and iterate over them
	  RecordDescriptor desc=stdfTypes.get("MRR");
      RecordData data = new RecordData(desc, mrrValues);
      Record record = new Record(desc,data );
      records.add(record);
  }
  
  private void generateWir() throws SQLException{
	  Object[] fieldValues =  null;
	  RecordDescriptor desc=stdfTypes.get("WIR");
	  RecordDescriptor wrrDesc=stdfTypes.get("WRR");
	  ResultSet wirs  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='WAFER_EVENT';");
	  while(wirs.next()){
		  if(fieldValues == null){
			  fieldValues =  new Object[(desc).size()];
		      fieldValues[desc.getIndex("SITE_GRP")] = 255;
		      fieldValues[desc.getIndex("HEAD_NUM")] = 1;
			  wrrValues =  new Object[(wrrDesc).size()];
		      wrrValues[wrrDesc.getIndex("SITE_GRP")] = 255;
		      wrrValues[wrrDesc.getIndex("HEAD_NUM")] = 1;
		  }
      	String attrName = wirs.getString(1);
      	Object attrValue = wirs.getObject(2);
        if( attrName.equals("WAFER_START_DTTM"))fieldValues[desc.getIndex("START_T")] = (Long) convertTimeToUnix((String)attrValue);
        if( attrName.equals("WAFER_STOP_DTTM"))wrrValues[wrrDesc.getIndex("FINISH_T")] = (Long) convertTimeToUnix((String)attrValue);
        if( attrName.equals("WAFER_FRAME_ID"))wrrValues[wrrDesc.getIndex("FRAME_ID")] = attrValue;
        if( attrName.equals("WAFER_MASK_ID"))wrrValues[wrrDesc.getIndex("MASK_ID")] = attrValue;
        if( attrName.equals("WAFER_USER_DESC"))wrrValues[wrrDesc.getIndex("USR_DESC")] = attrValue;
        if( attrName.equals("WAFER_EXEC_DESC"))wrrValues[wrrDesc.getIndex("EXC_DESC")] = attrValue;
	      
	      // get the wrr info now as well
		  ResultSet wrrs  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='WAFER_SUMMARY';");
		  while(wrrs.next()){
		        if( attrName.equals("PART_COUNT"))wrrValues[wrrDesc.getIndex("PART_CNT")] = attrValue;
		        if( attrName.equals("RETEST_PART_COUNT"))wrrValues[wrrDesc.getIndex("RTST_CNT")] = attrValue;
		        if( attrName.equals("ABORT_PART_COUNT"))wrrValues[wrrDesc.getIndex("RTST_CNT")] = attrValue;
		        if( attrName.equals("GOOD_PART_COUNT")){
		        	wrrValues[wrrDesc.getIndex("GOOD_CNT")] = attrValue;
		        	wrrValues[wrrDesc.getIndex("FUNC_CNT")] = attrValue;
		        }
		        if( attrName.equals("WAFER_EVENT_EID")){
		        	wrrValues[wrrDesc.getIndex("WAFER_ID")] = attrValue;
		        	fieldValues[desc.getIndex("WAFER_ID")] = attrValue;
		        }
			  }		  
	  }
	  if(fieldValues != null){
		  records.add(new Record(desc,(new RecordData(desc, fieldValues))));
	  }
  }
  
  private void generateWrr() throws SQLException{
	  if(mrrValues == null){
		  RecordDescriptor desc=stdfTypes.get("WRR");
		  records.add(new Record(desc,(new RecordData(desc, wrrValues))));
	  }
  }
  
  private void generateWcr() throws SQLException{
	  Object[] fieldValues =  null;
	  RecordDescriptor desc=stdfTypes.get("WCR");
	  ResultSet wcrs  = _sqlConnection.createStatement().executeQuery("select entityID from narrow where value='WAFER_INFO';");
	  while(wcrs.next()){
		  if(fieldValues == null) fieldValues =  new Object[(desc).size()];
	      String attrName = wcrs.getString(1);
	      Object attrValue = wcrs.getObject(2);
	      if( attrName.equals("WAFER_SIZE"))fieldValues[desc.getIndex("WAFR_SIZ")] = attrValue;
	      if( attrName.equals("DIE_HEIGHT"))fieldValues[desc.getIndex("DIE_HT")] = attrValue;
	      if( attrName.equals("DIE_HEIGHT"))fieldValues[desc.getIndex("DIE_WID")] = attrValue;
	      if( attrName.equals("WAFER_FLAT"))fieldValues[desc.getIndex("WAFER_UNITS")] = attrValue;
	      if( attrName.equals("CENTER_X"))fieldValues[desc.getIndex("CENTER_X")] = attrValue;
	      if( attrName.equals("CENTER_Y"))fieldValues[desc.getIndex("CENTER_Y")] = attrValue;
	      if( attrName.equals("POSITIVE_X_DIRECTION"))fieldValues[desc.getIndex("POS_X")] = attrValue;	      
	      if( attrName.equals("POSITIVE_Y_DIRECTION"))fieldValues[desc.getIndex("POS_Y")] = attrValue;
	      if( attrName.equals("WAFER_UNITS")){
	    	  if(attrValue.equals("NONE"))fieldValues[desc.getIndex("WF_UNITS")] = 0L;
	    	  if(attrValue.equals("in"))fieldValues[desc.getIndex("WF_UNITS")] = 1L;
	    	  if(attrValue.equals("cm"))fieldValues[desc.getIndex("WF_UNITS")] = 2L;
	    	  if(attrValue.equals("mm"))fieldValues[desc.getIndex("WF_UNITS")] = 3L;
	    	  if(attrValue.equals("mil"))fieldValues[desc.getIndex("WF_UNITS")] = 4L;
	      }		  
	  }
	  if(fieldValues != null) records.add(new Record(desc,(new RecordData(desc, fieldValues))));
  }
  
}
