package ri.stdf;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;

import ri.util.Emit;

import com.tragicphantom.stdf.Header;
import com.tragicphantom.stdf.Record;
import com.tragicphantom.stdf.RecordData;
import com.tragicphantom.stdf.RecordVisitor;
import com.tragicphantom.stdf.STDFReader;

/**This version of the STDF 4 to STDF 5 translator uses an XML mapping file*/
public class Stdf4To5TranslatorXml implements RecordVisitor {
  private String fileName=null;
  private int sequence=0;
  private int entityID=1; // runID is 1 first non run is 2
  private Connection _sqlConnection=null;          //database connection
  private Statement _sqlStatement=null;            //reusable statement for communicating with the database
  private PreparedStatement _stringInsert=null;    //resuable prepared statement
  //private PreparedStatement _intInsert=null;       //resuable prepared statement
  //private PreparedStatement _doubleInsert=null;    //resuable prepared statement
  
  private int runID=0;                             //entity ID for the run
  private int waferID=0;                           //entity ID for the current wafer
  
  private int[] currentTE=new int[256];            //entity ID per site for current test event, between PIR/PRR
  private Hashtable<Integer, Integer> testMap=new Hashtable<Integer, Integer>();    //a map from the test number in stdf to the entityID of the test
  private Hashtable<Integer, Boolean> testUpdate=new Hashtable<Integer, Boolean>(); //a map from testID to if an true if a PTR has been processed
  
  private static StdfMapping _stdfMapping = new StdfMapping(); //STDF4 to STDF5 mapping, read from XML file

  
  //@Override
  public void afterFile() {
  }

  //@Override
  public void beforeFile() {
  }

  /**Constructor*/
  public Stdf4To5TranslatorXml(File file, Connection sqlConnection) throws SQLException, FileNotFoundException, IOException, ParseException {
    fileName=file.getName();

    _sqlConnection=sqlConnection;
    _sqlStatement=_sqlConnection.createStatement(); //reusable statement for communicating with the database

    try { _sqlStatement.executeUpdate("PRAGMA synchronous = OFF;"); }  //speeds things up in sqlite by allowing the database to perform ops without verifyng the disk writes went through
    catch(Exception e) {}
    
    createTable("narrow", "narrow");
    _sqlConnection.setAutoCommit(false);
    STDFReader reader=new STDFReader(file); //note: accepts File, InputStream or filename
    reader.parse(this); //as it parses, the reader calls handleRecord (below) for each entry
    _sqlConnection.commit(); //commit all those changes
    _sqlConnection.setAutoCommit(true);

    try { _sqlStatement.executeUpdate("PRAGMA synchronous = ON;"); } //restore setting allowing the database to verify writes to the file system
    catch(Exception e) {}
  }
  
  /**Create the database and the single table
   * table is sequence int,entity int,index int, name string, value string | int | float
   * index is nullable, sequence is unique
   */
  private void createTable(String dName, String tName) {
    String tableName=tName; //Create database table if needed
    tableName=tName; //create sql table
    try {
      _sqlStatement.executeUpdate("create table if not exists " + tableName + " (" + 
          "sequence INTEGER PRIMARY KEY," + "entityID INTEGER," + "indexID INTEGER," + 
          "name TEXT," + "value NONE," + "value2 NONE);");

      _stringInsert=_sqlConnection.prepareStatement("insert into " + tableName + " values (?,?,?,?,?,?);"); //create a prepared sql statement that has the necessary fields for this stdf record type
    }
    catch(SQLException e) { e.printStackTrace(); }   
  }

  /**Three possible insert types,  no updates only inserts.
   * Populate the prepared statement and send it to the database
   * object can be String, Long, Double
   * */
  private boolean insertObject(int seq, int eid, int index, String name, Object val) {
    
    if(val == null) { return false; } //is a narrow table to ignore non-entries
    PreparedStatement prep=_stringInsert;
    sequence++;
    try {
      prep.setInt(1, sequence);
      prep.setInt(2, eid);
      prep.setInt(3, index);
      prep.setString(4, name);

      Class cl = val.getClass(); 
      if     (cl == String.class)  prep.setString(5, (String)val);
      else if(cl == Integer.class) prep.setInt(5, (Integer)val);
      else if(cl == Double.class)  prep.setDouble(5, (Double)val);
      else if(cl == Float.class)   prep.setFloat(5, (Float)val);
      
      prep.addBatch();
      prep.executeBatch();
    }
    catch(Exception e) {
      String emsg="Error in " + getClass().getName() + ".insertObject(): " + e;
      Emit.out(emsg);
      e.printStackTrace();
    }
    return true;
  }

  /** */
  private int getTestEntityFor(int testNumber) {
    if(testMap.containsKey(testNumber)) {
      return testMap.get(testNumber);
    }
    else {
      entityID++;
      testMap.put(testNumber, entityID);
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", "TEST_INFO");
      insertObject(sequence, entityID, 0, "TEST_NUM", testNumber); // put in the test number
      return entityID;
    }
  }

  /**Called by STDF parser for each record that comes in. 
   *@Override*/
  public void handleRecord(Record record) {
    try {
      RecordData recordData=record.getData(); //throws ParseException
      String recordType=recordData.getType(); //type of record (Mir, Ptr etc.)
      HashMap<String, Object> fields=recordData.getFields();
      int type=StdfUtil.getTypeFor(recordType);
      switch(type) {
        case StdfUtil.CNR: Emit.out("CNR not supported"); break;  //Cell Name Record
        case StdfUtil.FAR: Emit.out("FAR not supported"); break;  //File Attributes Record
        case StdfUtil.FTR: Emit.out("FTR not supported"); break;  //Functional Test Record
        case StdfUtil.MPR: Emit.out("MPR not supported"); break;  //Multiple-Result Parametric Record
        case StdfUtil.NMR: Emit.out("NMR not supported"); break;  //Name Map Record
        case StdfUtil.PLR: Emit.out("PLR not supported"); break;  //Pin List Record
        case StdfUtil.PSR: Emit.out("PSR not supported"); break;  //Pattern Sequence Record
        case StdfUtil.SCR: Emit.out("SCR not supported"); break;  //Scan Chain Description Record
        case StdfUtil.SSR: Emit.out("SSR not supported"); break;  //Scan Structure Record
        case StdfUtil.STR: Emit.out("STR not supported"); break;  //Scan Test Record
        case StdfUtil.VUR: Emit.out("VUR not supported"); break;  //Version Update Record
        default:
          insert(recordType, fields); //insert Mir, Bps, Dtr, Eps, etc.
          break;
      }
    }
    catch(ParseException e) {
      String emsg="Error in " + getClass().getName() + ".handleRecord(): " + e;
      Emit.out(emsg);
      System.out.print(emsg);
      //System.exit(1); //debug
    }
  }
  
  /**Handles Record mapping given a type (e.g. "Mir", "Wir", "Mrr", ...) */
  private void insert(String type, HashMap fields) {
    entityID++; // get the next entity ID
    runID=entityID;
    HashMap<String, String> mapto5 = _stdfMapping.getMapFor(type);

    String entityType = mapto5.get("ENTITY_TYPE"); //e.g. mir="RUN_INFO"
    if(entityType!=null)
      insertObject(sequence, entityID, 0, "ENTITY_TYPE", entityType);
    
    Iterator<String> iterFieldNames=fields.keySet().iterator();
    while(iterFieldNames.hasNext()) {
      String newName=null;
      Object fieldValue=null;
      String fieldName=(String)iterFieldNames.next();
      fieldValue=fields.get(fieldName); //set to null to drop
      newName = mapto5.get(fieldName);
      if(newName == null) {
        System.out.println(fieldName + " no mapping in mir");
        newName=fieldName;
      }
      if(fieldValue != null) {
        String val = fieldValue.toString().trim();
        if(!val.isEmpty()) 
          insertObject(sequence, runID, 0, newName, val);
      }
    }
  }

  @Override
  public void handleUnknownRecord(Header header, byte[] bytes) {
    Emit.out("WARNING: Skipping unknown record type: " + header.getType() + ", " + header.getSubType() + " [" + header.getLength());
    
  }

}
