package ri.stdf;

/**RI Constants and utilities for working with STDF.
 * 
 * Mainly this is name conversions: 
 *           long name <--> short name <--> database table name
 * 
 * Notes:
 * ----------------------------------------------------------------------------------------
 * 
 *     STDF Major Record Types
 *     -----------------------
 *      
 *       Major    Description
 *       -----    ---------------------------------------------
 *         0*     Information about the STDF file 
 *         1*     Data collected on a per lot basis 
 *         2      Data collected per wafer 
 *         5      Data collected on a per part basis 
 *        10      Data collected per test in the test program 
 *        15*     Data collected per test execution 
 *        20      Data collected per program segment 
 *        50      Generic Data
 *        
 *        
 *     STDF Record Structure
 *     --------------------- 
 *     The following table lists the meaning of the REC_TYP codes in STDF V4-2007. 
 *     
 *     REC_TYP  Code Meaning and STDF REC_SUB Codes
 *     -------  ------------     ------------------ 
 *
 *     0   Information about the STDF file 
 *     -------------------------------------------------- 
 *     10       File Attributes Record            (FAR)
 *     20       Audit Trail Record                (ATR)
 *     30       Version Update Record             (VUR)
 *
 *     1   Data collected on a per lot basis 
 *     -------------------------------------------------- 
 *     10       Master Information Record         (MIR)
 *     20       Master Results Record             (MRR)
 *     30       Part Count Record                 (PCR)
 *     40       Hardware Bin Record               (HBR)
 *     50       Software Bin Record               (SBR)
 *     60       Pin Map Record                    (PMR)
 *     62       Pin Group Record                  (PGR)
 *     63       Pin List Record                   (PLR)
 *     70       Retest Data Record                (RDR)
 *     80       Site Description Record           (SDR)
 *     90       Pattern Sequence Record           (PSR)
 *     91       Name Map Record                   (NMR)
 *     92       Cell Name Record                  (CNR)
 *     93       Scan Structure Record             (SSR)
 *     94       Scan Chain Record                 (SCR)
 *
 *     2   Data collected per wafer 
 *     -------------------------------------------------- 
 *     10       Wafer Information Record          (WIR)
 *     20       Wafer Results Record              (WRR)
 *     30       Wafer Configuration Record        (WCR)
 *     
 *     5   Data collected on a per part basis 
 *     -------------------------------------------------- 
 *     10       Part Information Record           (PIR)
 *     20       Part Results Record               (PRR)
 *     
 *     10   Data collected per test in the test program 
 *     -------------------------------------------------- 
 *     30       Test Synopsis Record              (TSR)
 *
 *     15  Data collected per test execution 
 *     -------------------------------------------------- 
 *     10       Parametric Test Record            (PTR)
 *     15       Multiple-Result Parametric Record (MPR)
 *     20       Functional Test Record            (FTR)
 *     30       Scan Test Record                  (STR)
 *
 *     20   Data collected per program segment 
 *     -------------------------------------------------- 
 *     10       Begin Program Section Record      (BPS)
 *     20       End Program Section Record        (EPS)
 *
 *     50   Generic Data
 *     -------------------------------------------------- 
 *     10       Generic Data Record               (GDR)
 *     30       Datalog Text Record               (DTR)
 *     
 * */
public class StdfUtil {

  public static final int UNK =  0;  //unknown record
  public static final int ATR =  1;  //Audit Trail Record
  public static final int BPS =  2;  //Begin Program Section Record
  public static final int CNR =  3;  //Cell Name Record
  public static final int DTR =  4;  //Datalog Text Record
  public static final int EPS =  5;  //End Program Section Record
  public static final int FAR =  6;  //File Attributes Record
  public static final int FTR =  7;  //Functional Test Record
  public static final int GDR =  8;  //Generic Data Record
  public static final int HBR =  9;  //Hardware Bin Record
  public static final int MIR = 10;  //Master Information Record
  public static final int MPR = 11;  //Multiple-Result Parametric Record
  public static final int MRR = 12;  //Master Results Record
  public static final int NMR = 13;  //Name Map Record
  public static final int PCR = 14;  //Part Count Record
  public static final int PGR = 15;  //Pin Group Record
  public static final int PIR = 16;  //Part Information Record
  public static final int PLR = 17;  //Pin List Record
  public static final int PMR = 18;  //Pin Map Record
  public static final int PRR = 19;  //Part Results Record
  public static final int PSR = 20;  //Pattern Sequence Record
  public static final int PTR = 21;  //Parametric Test Record
  public static final int RDR = 22;  //Retest Data Record
  public static final int SBR = 23;  //Software Bin Record
  public static final int SCR = 24;  //Scan Chain Description Record
  public static final int SDR = 25;  //Site Description Record
  public static final int SSR = 26;  //Scan Structure Record
  public static final int STR = 27;  //Scan Test Record
  public static final int TSR = 28;  //Test Synopsis Record
  public static final int VUR = 29;  //Version Update Record
  public static final int WCR = 30;  //Wafer Configuration Record
  public static final int WIR = 31;  //Wafer Information Record
  public static final int WRR = 32;  //Wafer Result Record
  
  /**Given "Hardware Bin Record" this returns "HBR", and so on
   * (note: can actually supply either the short name, long name, or dbase name)*/
  public static String getShortName(String name) {
    return getShortName(getTypeFor(name));
  }
  
  /**Given the type (HBR, MIR, etc.) this returns the short name, i.e. "HBR"*/
  public static String getShortName(int type) {
    switch(type) {
      case ATR: return "ATR";
      case BPS: return "BPS";
      case CNR: return "CNR";
      case DTR: return "DTR";
      case EPS: return "EPS";
      case FAR: return "FAR";
      case FTR: return "FTR";
      case GDR: return "GDR";
      case HBR: return "HBR";
      case MIR: return "MIR";
      case MPR: return "MPR";
      case MRR: return "MRR";
      case NMR: return "NMR";
      case PCR: return "PCR";
      case PGR: return "PGR";
      case PIR: return "PIR";
      case PLR: return "PLR";
      case PMR: return "PMR";
      case PRR: return "PRR";
      case PSR: return "PSR";
      case PTR: return "PTR";
      case RDR: return "RDR";
      case SBR: return "SBR";
      case SCR: return "SCR";
      case SDR: return "SDR";
      case SSR: return "SSR";
      case STR: return "STR";
      case TSR: return "TSR";
      case VUR: return "VUR";
      case WCR: return "WCR";
      case WIR: return "WIR";
      case WRR: return "WRR";
      default:  return null; //unknown record type
    }
  }

  /**Given "HBR" this returns "Hardware Bin Record", and so on
   * (note: can actually supply either the short name, long name, or dbase name)*/
  public static String getLongName(String name) {
    return getLongName(getTypeFor(name));
  }
  
  /**Given the type (HBR, MIR, etc.) this returns the long name, i.e. "Hardware Bin Record"*/
  public static String getLongName(int type) {
    switch(type) {
      case ATR: return "Audit Trail Record";
      case BPS: return "Begin Program Section Record";
      case CNR: return "Cell Name Record";
      case DTR: return "Datalog Text Record";
      case EPS: return "End Program Section Record";
      case FAR: return "File Attributes Record";
      case FTR: return "Functional Test Record";
      case GDR: return "Generic Data Record";
      case HBR: return "Hardware Bin Record";
      case MIR: return "Master Information Record";
      case MPR: return "Multiple-Result Parametric Record";
      case MRR: return "Master Results Record";
      case NMR: return "Name Map Record";
      case PCR: return "Part Count Record";
      case PGR: return "Pin Group Record";
      case PIR: return "Part Information Record";
      case PLR: return "Pin List Record";
      case PMR: return "Pin Map Record";
      case PRR: return "Part Results Record";
      case PSR: return "Pattern Sequence Record";
      case PTR: return "Parametric Test Record";
      case RDR: return "Retest Data Record";
      case SBR: return "Software Bin Record";
      case SCR: return "Scan Chain Description Record";
      case SDR: return "Site Description Record";
      case SSR: return "Scan Structure Record";
      case STR: return "Scan Test Record";
      case TSR: return "Test Synopsis Record";
      case VUR: return "Version Update Record";
      case WCR: return "Wafer Configuration Record";
      case WIR: return "Wafer Information Record";
      case WRR: return "Wafer Result Record";
      default:  return null; //unknown record type
    }
  }

  /**This is the name used for database tables.
   * Given "HBR" or "Hardware Bin Record" this returns "HARDWARE_BIN", and so on
   * (note: can actually supply either the short name, long name, or dbase name)*/
  public static String getDbaseName(String name) {
    return getDbaseName(getTypeFor(name));  
  }
  
  /**This is the name used for database tables.
   * Given the type (HBR, MIR, etc) this returns the databse table name, i.e. "HARDWARE_BIN"*/
  public static String getDbaseName(int type) {
    switch(type) {
      case ATR: return "AUDIT_TRAIL";
      case BPS: return "BEGIN_PROGRAM_SECTION";
      case CNR: return "CELL_NAME";
      case DTR: return "DATALOG_TEXT";
      case EPS: return "END_PROGRAM_SECTION";
      case FAR: return "FILE_ATTRIBUTES";
      case FTR: return "FUNCTIONAL_TEST";
      case GDR: return "GENERIC_DATA";
      case HBR: return "HARDWARE_BIN";
      case MIR: return "MASTER_INFORMATION";
      case MPR: return "MULTIPLE_RESULT_PARAMETRIC";
      case MRR: return "MASTER_RESULTS";
      case NMR: return "NAME_MAP";
      case PCR: return "PART_COUNT";
      case PGR: return "PIN_GROUP";
      case PIR: return "PART_INFORMATION";
      case PLR: return "PIN_LIST";
      case PMR: return "PIN_MAP";
      case PRR: return "PART_RESULTS";
      case PSR: return "PATTERN_SEQUENCE";
      case PTR: return "PARAMETRIC_TEST";
      case RDR: return "RETEST_DATA";
      case SBR: return "SOFTWARE_BIN";
      case SCR: return "SCAN_CHAIN_DESCRIPTION";
      case SDR: return "SITE_DESCRIPTION";
      case SSR: return "SCAN_STRUCTURE";
      case STR: return "SCAN_TEST";
      case TSR: return "TEST_SYNOPSIS";
      case VUR: return "VERSION_UPDATE";
      case WCR: return "WAFER_CONFIGURATION";
      case WIR: return "WAFER_INFORMATION";
      case WRR: return "WAFER_RESULT";
      default:  return null; //unknown record type
    }
  }

  /**Returns the record type (as an int) given a short name, long name, or dbase table name.
   * if given "PRT" or "Parametric Test Record" or "PARAMETRIC_TEST" then returns the corresponding int   
   */
  public static int getTypeFor(String name) {
    if(name==null || name.length()<3) return UNK; //unknown record type
    
    switch(name.charAt(0)) {
      
      case 'A': case 'a':
        if(name.equalsIgnoreCase("ATR"))                return ATR; 
        if(name.equalsIgnoreCase("Audit Trail Record")) return ATR;
        if(name.equalsIgnoreCase("AUDIT_TRAIL"))        return ATR; 
        break;
      
      case 'B': case 'b':
        if(name.equalsIgnoreCase("BPS"))                          return BPS; 
        if(name.equalsIgnoreCase("Begin Program Section Record")) return BPS;
        if(name.equalsIgnoreCase("BEGIN_PROGRAM_SECTION"))        return BPS ; 
        break;
      
      case 'C': case 'c':
        if(name.equalsIgnoreCase("CNR"))              return CNR; 
        if(name.equalsIgnoreCase("Cell Name Record")) return CNR;
        if(name.equalsIgnoreCase("CELL_NAME"))        return CNR; 
        break;
      
      case 'D': case 'd':
        if(name.equalsIgnoreCase("DTR"))                 return DTR; 
        if(name.equalsIgnoreCase("Datalog Text Record")) return DTR;
        if(name.equalsIgnoreCase("DATALOG_TEXT"))        return DTR; 
        break;
      
      case 'E': case 'e':
        if(name.equalsIgnoreCase("EPS"))                        return EPS; 
        if(name.equalsIgnoreCase("End Program Section Record")) return EPS;
        if(name.equalsIgnoreCase("END_PROGRAM_SECTION"))        return EPS; 
        break;
      
      case 'F': case 'f':
        if(name.equalsIgnoreCase("FAR")) return FAR; 
        if(name.equalsIgnoreCase("FTR")) return FTR; 

        if(name.equalsIgnoreCase("File Attributes Record")) return FAR;
        if(name.equalsIgnoreCase("Functional Test Record")) return FTR;
        
        if(name.equalsIgnoreCase("FILE_ATTRIBUTES")) return FAR; 
        if(name.equalsIgnoreCase("FUNCTIONAL_TEST")) return FTR; 
        break;
      
      case 'G': case 'g':
        if(name.equalsIgnoreCase("GDR"))                 return GDR; 
        if(name.equalsIgnoreCase("Generic Data Record")) return GDR;
        if(name.equalsIgnoreCase("GENERIC_DATA"))        return GDR; 
        break;
      
      case 'H': case 'h':
        if(name.equalsIgnoreCase("HBR"))                 return HBR; 
        if(name.equalsIgnoreCase("Hardware Bin Record")) return HBR;
        if(name.equalsIgnoreCase("HARDWARE_BIN"))        return HBR; 
        break;
      
      case 'M': case 'm':
        if(name.equalsIgnoreCase("MIR")) return MIR; 
        if(name.equalsIgnoreCase("MPR")) return MPR; 
        if(name.equalsIgnoreCase("MRR")) return MRR; 

        if(name.equalsIgnoreCase("Master Information Record"))         return MIR;
        if(name.equalsIgnoreCase("Multiple-Result Parametric Record")) return MPR;
        if(name.equalsIgnoreCase("Master Results Record"))             return MRR;
        
        if(name.equalsIgnoreCase("MASTER_INFORMATION"))         return MIR; 
        if(name.equalsIgnoreCase("MULTIPLE_RESULT_PARAMETRIC")) return MPR; 
        if(name.equalsIgnoreCase("MASTER_RESULTS"))             return MRR; 
        break;
      
      case 'N': case 'n':
        if(name.equalsIgnoreCase("NMR"))             return NMR; 
        if(name.equalsIgnoreCase("Name Map Record")) return NMR;
        if(name.equalsIgnoreCase("NAME_MAP"))        return NMR; 
        break;
      
      case 'P': case 'p':
        if(name.equalsIgnoreCase("PCR")) return PCR; 
        if(name.equalsIgnoreCase("PGR")) return PGR; 
        if(name.equalsIgnoreCase("PIR")) return PIR; 
        if(name.equalsIgnoreCase("PLR")) return PLR; 
        if(name.equalsIgnoreCase("PMR")) return PMR; 
        if(name.equalsIgnoreCase("PRR")) return PRR; 
        if(name.equalsIgnoreCase("PSR")) return PSR; 
        if(name.equalsIgnoreCase("PTR")) return PTR; 
        
        if(name.equalsIgnoreCase("Part Count Record"))       return PCR;
        if(name.equalsIgnoreCase("Pin Group Record"))        return PGR;
        if(name.equalsIgnoreCase("Part Information Record")) return PIR;
        if(name.equalsIgnoreCase("Pin List Record"))         return PLR;
        if(name.equalsIgnoreCase("Pin Map Record"))          return PMR;
        if(name.equalsIgnoreCase("Part Results Record"))     return PRR;
        if(name.equalsIgnoreCase("Pattern Sequence Record")) return PSR;
        if(name.equalsIgnoreCase("Parametric Test Record"))  return PTR;

        if(name.equalsIgnoreCase("PART_COUNT"))       return PCR; 
        if(name.equalsIgnoreCase("PIN_GROUP"))        return PGR; 
        if(name.equalsIgnoreCase("PART_INFORMATION")) return PIR; 
        if(name.equalsIgnoreCase("PIN_LIST"))         return PLR; 
        if(name.equalsIgnoreCase("PIN_MAP"))          return PMR; 
        if(name.equalsIgnoreCase("PART_RESULTS"))     return PRR; 
        if(name.equalsIgnoreCase("PATTERN_SEQUENCE")) return PSR; 
        if(name.equalsIgnoreCase("PARAMETRIC_TEST"))  return PTR;
        break;
      
      case 'R': case 'r':
        if(name.equalsIgnoreCase("RDR"))                return RDR; 
        if(name.equalsIgnoreCase("Retest Data Record")) return RDR;
        if(name.equalsIgnoreCase("RETEST_DATA"))        return RDR; 
        break;
      
      case 'S': case 's':
        if(name.equalsIgnoreCase("SBR")) return SBR; 
        if(name.equalsIgnoreCase("SCR")) return SCR; 
        if(name.equalsIgnoreCase("SDR")) return SDR; 
        if(name.equalsIgnoreCase("SSR")) return SSR; 
        if(name.equalsIgnoreCase("STR")) return STR; 

        if(name.equalsIgnoreCase("Software Bin Record"))           return SBR;
        if(name.equalsIgnoreCase("Scan Chain Description Record")) return SCR;
        if(name.equalsIgnoreCase("Site Description Record"))       return SDR;
        if(name.equalsIgnoreCase("Scan Structure Record"))         return SSR;
        if(name.equalsIgnoreCase("Scan Test Record"))              return STR;
        if(name.equalsIgnoreCase("Scan Cell Name Record"))         return CNR; //just in case (the real name is "Cell Name Record")

        if(name.equalsIgnoreCase("SOFTWARE_BIN"))           return SBR; 
        if(name.equalsIgnoreCase("SCAN_CHAIN_DESCRIPTION")) return SCR; 
        if(name.equalsIgnoreCase("SITE_DESCRIPTION"))       return SDR; 
        if(name.equalsIgnoreCase("SCAN_STRUCTURE"))         return SSR; 
        if(name.equalsIgnoreCase("SCAN_TEST"))              return STR; 
        break;
      
      case 'T': case 't':
        if(name.equalsIgnoreCase("TSR"))                  return TSR; 
        if(name.equalsIgnoreCase("Test Synopsis Record")) return TSR;
        if(name.equalsIgnoreCase("TEST_SYNOPSIS"))        return TSR; 
        break;
      
      case 'V': case 'v':
        if(name.equalsIgnoreCase("VUR"))                   return VUR; 
        if(name.equalsIgnoreCase("Version Update Record")) return VUR;
        if(name.equalsIgnoreCase("VERSION_UPDATE"))        return VUR; 
        break;
      
      case 'W': case 'w':
        if(name.equalsIgnoreCase("WCR")) return WCR; 
        if(name.equalsIgnoreCase("WIR")) return WIR; 
        if(name.equalsIgnoreCase("WRR")) return WRR; 

        if(name.equalsIgnoreCase("Wafer Configuration Record")) return WCR;
        if(name.equalsIgnoreCase("Wafer Information Record"))   return WIR;
        if(name.equalsIgnoreCase("Wafer Result Record"))        return WRR;

        if(name.equalsIgnoreCase("WAFER_CONFIGURATION")) return WCR; 
        if(name.equalsIgnoreCase("WAFER_INFORMATION"))   return WIR; 
        if(name.equalsIgnoreCase("WAFER_RESULT"))        return WRR; 
        break;
    }
    return UNK; //catch-all for unknown record types
  }

}
