package ri.util;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;

import javax.swing.JList;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.Timer;
import javax.swing.border.BevelBorder;
import javax.swing.border.CompoundBorder;
import javax.swing.border.LineBorder;


/** A General Purpose JList Dialog that initiates an action when clicked on.
 *  Uses the java.awt.Window as its base class so no Frame is drawn.
 *  This is a convenient action event oriented list
 *  that looks and operates similar to the little dialog that comes
 *  up on the right click button in Windows and Netscape.
 * */
public class ActionList extends Window {
  private JList _jlist;
  private ActionListListener _actn;
  private MouseEvent _sourceEvent; //Mouse event that caused this window to pop up
  private boolean _disposeOnExit= true; //If true then window disposes self on exit
  private final static ActionEvent _actionEvent= new ActionEvent(new Object(), 0, "ActionList Item Selected"); //For reuse
  private Timer _exitTimer = new Timer(1000, new ActionListener() { //closes the window when it times out
    public void actionPerformed(ActionEvent e) {
      doExit();      
    }
  });

  /**Constr:
   * Owner is REQUIRED for the base class and may not be null.
   * Owner and mouseEvent are used to position the window.
   * ActionListListener provides the values to be displayed as well as
   * the actions to be performed*/
  public ActionList(Component owner, MouseEvent sourceEvent, ActionListListener actn) {
    super(getWindowFor(owner));
    _actn= actn;
    _sourceEvent = sourceEvent;
    init();
    setPositionFor(owner, sourceEvent);
  }
  
  private void doExit() {
    if (_disposeOnExit) dispose();
    else setVisible(false);
  }
  
  /**Used to override default display of how the action entries are displayed*/
  public void setRenderer(ListCellRenderer cellRenderer) {
    _jlist.setCellRenderer(cellRenderer);
  }
  
  /**Sets the tool tip text of the contained list*/
  public void setToolTipText(String s) {
    if(_jlist!=null) _jlist.setToolTipText(s);
  }

  /**Used by constr to initialize this object*/
  private void init() {
    _jlist= new JList(_actn.getActions());
    _jlist.setBorder(new CompoundBorder(new LineBorder(Color.gray), new BevelBorder(BevelBorder.RAISED)));
    _jlist.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    add(_jlist, BorderLayout.CENTER);

    //Performs action or else cancels (if mouse leaves window boundries):
    _jlist.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        Object ob= _jlist.getSelectedValue();
        if (ob instanceof ActionListener)
           ((ActionListener)ob).actionPerformed(_actionEvent);
        ///else if(ob instanceof ActionList) //ie: a sub-menu
        ///  ((ActionList)ob).show();
        else {
          ///_actn.doAction(_jlist.getSelectedIndex(), e); //Passes mouse event in the context box
          _actn.doAction(_jlist.getSelectedIndex(), _sourceEvent); //original source event that caused the context menu to come up
        }
        doExit();
      }
      public void mouseExited(MouseEvent e) {
        _exitTimer.restart();
        //doExit();
      }
      public void mouseEntered(MouseEvent e) {
        if(_exitTimer.isRunning())
          _exitTimer.stop();
      }
    });

    //Highlights closest list entry: 
    _jlist.addMouseMotionListener(new MouseMotionAdapter() {
      public void mouseMoved(MouseEvent e) {
        _jlist.setSelectedIndex(_jlist.locationToIndex(e.getPoint()));
      }
    });

    //Exits on key: 'ESC'
    addKeyListener(new KeyAdapter() {
      public void keyTyped(KeyEvent e) {
        if (e.getKeyChar() == 27)
          doExit(); //Escape key
      }
    });

    /* ***Note: requires Java 1.4: *** * /
    addWindowFocusListener(new WindowFocusAdapter() {
      public void windowLostFocus(WindowEvent e) { doExit(); }
    });
    /* *** */

    pack();
  }

  /**If true then window disposes self on exit
   * (ie: can't be reused)*/
  public void setDisposeOnExit(boolean b) {
    _disposeOnExit= b;
  }

  /**Sets position of this window. Next to mouse click but not off screen*/
  public void setPositionFor(Component owner, MouseEvent e) {
    Dimension scnSize= Toolkit.getDefaultToolkit().getScreenSize();

    Point ownerPos;
    if (owner == null) //Just in case
      ownerPos= new Point(scnSize.width / 2, scnSize.height / 2);
    else ownerPos= owner.getLocation();

    //Find a parent of type java.awt.Window and adjust location relative to it:
    Window w= null;
    Component c= owner;
    while (w == null && c != null) {
      if (c instanceof Window)
        w= (Window)c;
      else {
        c= c.getParent();
        Point parentPos= c.getLocation();
        ownerPos.x += parentPos.x;
        ownerPos.y += parentPos.y;
      }
    }

    Point clickLoc;
    if (e == null)
      clickLoc= new Point(0, 0); //Just in case
    else clickLoc= e.getPoint();

    Dimension mySize= getSize();

    int xloc= ownerPos.x + clickLoc.x;
    int yloc= ownerPos.y + clickLoc.y;

    if ((xloc + mySize.width) > scnSize.width)
      xloc= scnSize.width - mySize.width;

    int bottomMargin= 30; //Assumed size of Windows taskbar
    if ((yloc + mySize.height) > scnSize.height - bottomMargin)
      yloc= scnSize.height - mySize.height - bottomMargin;

    if (xloc < 10) xloc= 10;
    if (yloc < 10) yloc= 10;

    setLocation(xloc, yloc);
  }

  public void setVisible(boolean b) {
    if(b) this.toFront(); //This helps
    super.setVisible(b);
  }

  public void show(Component owner, MouseEvent e) {
    setPositionFor(owner, e);
    setVisible(true);
  }
  
  /**Given a Component, This follows its parental lineage until it hits
   *anything that qualifies as a java.awt.Window and returns that.
   * Normally Windows are: Window, Dialog, and Frame components.
   */
  private static Window getWindowFor(Component c) {
    while (c != null) {
      if (c instanceof Window) return (Window)c;
      c = c.getParent();
    }
    return null;
  }

  
}
