package ri.util;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Iterator;

/** This is like a StringBuffer but for byte arrays */
public class ByteBuffer {

  private byte[] buf;
  private int curpos=0; // points at next location to next write pos
                        // when reading points at next read position

  /** Constr */
  public ByteBuffer() { buf=new byte[4000]; }

  /**Constr*/
  public ByteBuffer(int capacity) { buf=new byte[capacity]; }

  /**Constr*/
  public ByteBuffer(byte[] initialEntries) { buf=initialEntries; }

  /**Append to this buffer*/
  public void append(byte[] data) {
    int len=data == null ? 0 : data.length;
    append(data, 0, len);
  }

  /**Append to this buffer*/
  public void append(byte[] data, int loc, int cnt) {
    if((curpos + cnt) >= buf.length) {
      int newSize=buf.length * 2; //assumption, grow by factor of two, at least
      int needed=curpos + cnt;    //min size needed for the new data
      if(needed > newSize) newSize=(needed * 3) / 2;
      growTo(newSize);
    }
    for(int i=loc; i < cnt; i++)
      buf[curpos++]=data[i];
  }

  /**Append to this buffer*/
  public void append(byte b) {
    int curSize=buf.length;
    if((curpos + 1) >= curSize) growTo(curSize * 2); //double each time
    buf[curpos++]=b;
  }

  /**Removes specified number of entries from the end of this buffer.
   * No errors for specifying an illegal number.
   * Returns the actual number of entries removed*/
  public int removeFromEnd(int cnt) {
    curpos-=cnt;
    int actual;
    if(curpos < 0) {
      actual=cnt + curpos;
      curpos=0;
    }
    else actual=cnt;
    return actual;
  }

  /**Removes specified number of entries from the start of this buffer.
   * No errors for specifying an illegal number.
   * Returns the actual number of entries removed
   * TODO: Improve this one (very inefficient)***
   * */
  public int removeFromStart(int cnt) {
    int avail=size();
    if(cnt >= avail) {
      clear();
      return avail;
    }
    int newLen=avail - cnt;
    byte[] ba=new byte[newLen];
    getBytes(cnt, avail, ba, 0);
    clear(); //Inefficient to destroy the whole thing and copy back in***
    append(ba);
    return cnt;
  }

  /** Clears out the contents of the buffer */
  public void clear() { curpos=0; }

  /**Returns the current capacity of this buffer,
   * ie: how many entries it can hold without 'growing'*/
  public int capacity() { return buf.length; }

  /**Returns element at location i
   * Note: Throws ArrayIndexOutOfBounds on attempt to fetch illegal index*/
  public byte get(int loc) { return buf[loc]; }

  /**Get a subrange of entries, places it in 'dst' at 'dstBegin'
   *      inclusive of srcBegin, exclusive of srcEnd
   *      Won't overfill dst array.
   */
  public void getBytes(int srcBegin, int srcEnd, byte[] dst, int dstBegin) {
    int len=srcEnd - srcBegin;
    int dstLen=dst.length - dstBegin;
    if(dstLen < len) len=dstLen; //Just in case dst is too small
    int i=srcBegin;
    int j=dstBegin;
    int end=srcBegin + len;
    while(i < end) dst[j++]=buf[i++];
  }

  /** Grows to at least length minSize */
  synchronized private void growTo(int minSize) {
    int newLen=buf.length * 2;
    if(newLen < minSize) newLen=minSize;
    byte[] newBuf=new byte[newLen];
    int length=buf.length;
    for(int i=0; i < length; i++) newBuf[i]=buf[i];
    buf=newBuf;
  }

  /**Set buffer location to specified (byte) value*/
  public void set(int loc, int val) {
    if(loc >= curpos) growTo(loc + 10);
    buf[loc]=(byte)(val & 0xff);
  }

  /**Returns current number of entries in this buffer*/
  public int size() { return curpos; }

  /**Sets size to specified value.  If val, compared to current size, is:
   *   -Less: Truncates the buffer down to the specified size.
   *          (Note: does not reduce size of buffer)
   *   -Greater: Zeroes elements between current size and new one. Grows capacity as required.
   */
  public void size(int val) {
    if(val < 0) return;
    if(val < curpos) curpos=val; //truncate size
    if(capacity() < val) growTo(val);
    while(curpos < val) buf[curpos++]=0;
  }

  /**Returns this buffer as a byte array*/
  public byte[] toBytes() {
    if(curpos == buf.length) return buf;
    byte[] ba=new byte[curpos];
    for(int i=0; i < curpos; i++) ba[i]=buf[i];
    return ba;
  }

  /**Override: Returns this buffer as a String*/
  public String toString() { return new String(buf, 0, curpos); }
  
  /**
   * Add cr/lf
   * 
   */
  public void eol(){
    append((byte)0xd);
    append((byte)0xa);
  }
  /**
   * Add riri sep ^ + ~ `
   * 
   */
  public void sep(Character c){
    if( c == '^') append((byte)0x1c);
    if( c == '+') append((byte)0x1d);
    if( c == '~') append((byte)0x1e);
    if( c == '`') append((byte)0x1f);
  }
  /**
   * Convert string to bytes and append
   * @param text
   */
  public void append(String text){
    try {
      append(text.getBytes("US-ASCII"));
    }
    catch (UnsupportedEncodingException e) {
    }
  }
  public void appendLine(String text){
    append(text);
    eol();
  }
  /**
   * Append a string mapping riri seps to bytes ^ + ~ `
   * @param string
   */
  public void appendSmString(String string){
    // replaces ^ + ~ ' with1C 1D 1E 1F
    byte[] buffer = null;
    try {
      buffer=string.getBytes("US-ASCII");
    }
    catch (UnsupportedEncodingException e) {
    }
    for(int i = 0; i < buffer.length; i++ ){
      if(buffer[i] == 94) {
        append((byte)0x1c);
      } else if (buffer[i] == 43){
        append((byte)0x1d);        
      } else if (buffer[i] == 126){
        append((byte)0x1e);        
      }
      if(buffer[i] == 96){
        append((byte)0x1f);
      }else{
        append(buffer[i]);
      }
    }
  }
  
  /**Returns true if no more to process*/
  public boolean atEnd() {
    return curpos >= buf.length;
  }
  
  public  byte nextByte(){
    // get the byte at the cursor and advance it
    byte rtn = get(curpos);
    curpos++;
    return rtn;
  }
  
  public  byte peekByte(){
    // get the byte at the cursor
    byte rtn = get(curpos);
    return rtn;
  }
  
  public  void  putByte(byte val){
    // put the byte at the cursor
    buf[curpos] = val;
    curpos++;
  }
  
  // CBOR support
  /**Expects next value to be a tag and returns its value
   * throws an exception if not there*/
  public int nextTag() {
    byte tag = peekByte();
    if((tag & 0xe0) != 0xc0) return 0;
    int size = nextCborSize();
    return size;
  }

/** Computes the CBOR size of the next element*/
  public int nextCborSize() {
    int  tag = (int)nextByte() & 0x1f; 
    if( tag < 24) return (int)tag;
    if( tag == 24) return (int)nextByte() & 0xff;
    if( tag == 25){
      int tmp = ((int)nextByte() & 0xff) * 256;
      int tmp2 = (int)nextByte() & 0xff;
      return tmp + tmp2;
    }
    return 0;
  }
  
  /**Expects next value to be a string and returns its value
   * throws an exception if not there*/
  public String nextString() {
    byte tag = peekByte();
    if((tag & 0xe0) != 0x60) return "";
    int size = nextCborSize();
    byte[] rtn = new byte[size];
    getBytes(curpos, curpos + size, rtn, 0);
    curpos = curpos + size;
    try {
      return new String(rtn, "US-ASCII");
    }
    catch (UnsupportedEncodingException e) {
    }
    return "";
  }
  
  public byte[] nextBytes() {
    byte tag = peekByte();
    if((tag & 0xe0) != 0x40) return new byte[0];
    int size = nextCborSize();
    byte[] rtn = new byte[size];
    getBytes(curpos, curpos + size, rtn, 0);
    curpos = curpos + size;
    return rtn;
  }
  
 public void fillMap(HashMap map){
   // starting at the current location which defines a map
   // get key values and insert into the hashMap
   int size = nextCborSize(); // number of map elements
   for( int i = 0; i < size; i++){
     String key = nextString();
     String value = nextString();
     map.put(key,value);
   }
 }
 
 public void writeString(String value){
   int size = value.length();
   if(size < 24){
     putByte((byte) (0x60 + size));
   }else if(size < 256){
     putByte((byte) (0x60 + 24));
     putByte((byte) size);
   }else{
     putByte((byte) (0x60 + 25));
     putByte((byte) (size >> 8));
     putByte((byte) (size & 0xff));

   }
   append(value);
 }
 
 public void writeBytes(byte[] value){
   int size = value.length;
   if(size < 24){
     putByte((byte) (0x40 + size));
   }else if(size < 256){
     putByte((byte) (0x40 + 24));
     putByte((byte) size);
   }else{
     putByte((byte) (0x40 + 25));
     putByte((byte) (size >> 8));
     putByte((byte) (size & 0xff));
   }
   append(value);
 }
 
 public void writeTag(int tag){
   if(tag < 24){
     putByte((byte) (0xc0 + tag));
   }else if(tag < 256){
     putByte((byte) (0xc0 + 24));
     putByte((byte) tag);
   }else{
     putByte((byte) (0xc0 + 25));
     putByte((byte) (tag >> 8));
     putByte((byte) (tag & 0xff));
   }
 }
 
 public void writeArraySize(int size){
   if(size < 24){
     putByte((byte) (0x80 + size));
   }else if(size < 256){
     putByte((byte) (0x80 + 24));
     putByte((byte) size);
   }else{
     putByte((byte) (0x80 + 25));
     putByte((byte) (size >> 8));
     putByte((byte) (size & 0xff));
   }
 }
 
 public void writeMapSize(int size){
   if(size < 24){
     putByte((byte) (0xa0 + size));
   }else if(size < 256){
     putByte((byte) (0xa0 + 24));
     putByte((byte) size);
   }else{
     putByte((byte) (0xa0 + 25));
     putByte((byte) (size >> 8));
     putByte((byte) (size & 0xff));
   }
 }
 
 public void writeHashMap(HashMap map){
   int size = map.size();
   writeMapSize(size);
   Iterator<String> iter=map.keySet().iterator();
   while (iter.hasNext()) {
     String key=(String)iter.next(); //guru key name
     String val=(String)map.get(key); //value for key
     writeString(key);
     writeString(val);
   }
 }
}
