package ri.util;

import java.util.Collection;
import java.util.Iterator;


/**
 * FILE:    Csv.java
 *  Author:  O. Rodriguez
 *  Company: Roos Instruments Inc.
 *
 *  CSV (Comma Separated Variables) string class.
 *  Given a CSV string: Separates the input string into an array of Strings
 *     ie: aaa,bbb,"ccc,ddd" ==> String[] {"aaa", "bbb", "ccc,ddd"}
 *  Given an array of Strings: Combines them into a CSV string
 *     ie: String[] {"aaa", "bbb", "ccc,ddd"} ==> aaa,bbb,'ccc,ddd'
 *  Notes:
 *    -Commas within double quotes are not treated as field separators
 *    -There is no mechanism (currently) for including double quotes within files
 *    -Performs conversion using a 'just in time' approach, ie doesn't convert
 *     until is needed, caches the result.
 *
 *  Added toCsv(Collection) which recursively emits the collection using a hierarchy of separators
 *  History:
 *  5/7/1999 OER: moved to package ri.oer.rix.cff
 *  ~3/1998 OER: created
 *
 */
public class Csv {

  private String csvString=null;
  private String[] separatedList=null;
  private static final String[] StringA0=new String[0]; //Used for blank lines
  private boolean _recognizeEmbeddedQuotes=true;

  /**Constructor given an array of Strings (to be converted to CSV)*/
  public Csv(String[] sa) {
    separatedList=sa;
    csvString=null;
  }

  /**Constr given a CSV string (to be converted to String[]*/
  public Csv(String s) {
    csvString=s;
    separatedList=null;
  }

  /**Utility method. Combines the given string array into a String
   * using the given separator character.
   * Note: Does not return null.*/
  public final static String makeStringSepBy(String[] sa, char sepBy) {
    int len=sa == null ? 0 : sa.length;
    switch (len) {
      case 0: return "";
      case 1: return sa[0];
      default:
        StringBuffer sb=new StringBuffer(sa[0]);
        for (int i=1; i < len; i++)
          sb.append(sepBy).append(sa[i]);
        return sb.toString();
    }
  }

  /**Return this as a String[]*/
  public final String[] get() {
    if (separatedList == null) //Do lazy evaluation
      separatedList=toArray(csvString);
    return separatedList;
  }

  /**Returns this as a CSV (ie: comma separated) String*/
  public final String getCsv() {
    return getCsv(0);
  }

  /**Returns as CSV String, starting at the specified (zero based) index*/
  public final String getCsv(int startIndex) {
    get(); //Ensures that separatedList is up to date
    if (separatedList == null) return "";
    int stopIndex=separatedList.length - 1;
    return makeCsv(separatedList, startIndex, stopIndex);
  }

  /**Returns as CSV String, given the specified (zero based) indices.
   * Note: Both indices are inclusive.*/
  public final String getCsv(int startIndex, int stopIndex) {
    get(); //Ensures that separatedList is up to date
    if (separatedList == null) return "";
    int lastIndex=separatedList.length - 1;
    if (stopIndex > lastIndex) stopIndex=lastIndex;
    return makeCsv(separatedList, startIndex, stopIndex);
  }

  /** Returns a CSV string created from the internal String array */
  private final static String makeCsv(String[] sa, int startIndex, int stopIndex) {
    if (startIndex > stopIndex) return "";
    if ((sa.length == 1) && (sa[0].indexOf(',') < 0)) //Avoid creating StringBuffer if poss.
      return sa[0];

    StringBuffer sb=new StringBuffer();
    for (int i=startIndex; i <= stopIndex; i++) {
      String s=sa[i];
      if (s == null) sb.append("");
      else {
        if (s.indexOf(',') >= 0) //Put quotes on fields with commas
        sb.append('\"' + s + '\"');
        else sb.append(s);
      }
      if (i != stopIndex) sb.append(",");
    } //for
    return sb.toString();
  }

  /**Set value from String[] */
  public final void set(String[] sa) {
    csvString=null;
    separatedList=sa;
  }

  /**Set value from CSV String */
  public final void set(String s) {
    separatedList=null;
    csvString=s;
  }

  /**Performs the actual separation of a CSV String into a String[]
   Note: Does not return null
   */
  public final static String[] toArray(String s) {
    return toArray(s, ',');
  }

  /**Separates, into a String[], the given String using the specified separator character
     Note: Does not return null
  */
  public final static String[] toArray(String s, char sepBy) {
    if(s==null) return StringA0;
    int strLen=s.length();
    if(strLen==0) return StringA0; //ie: String[0]

    //First, Count the sepBy chars that are not between quotes so can allocate array:
    int fieldCnt=0;
    boolean inQuotes=false;

    for (int i=0; i < strLen; i++) {
      char ch=s.charAt(i);
      if (ch == '\"') inQuotes=!inQuotes;
      else if (!inQuotes && ch == sepBy) fieldCnt++;
    }

    fieldCnt++; //One more for good measure
    String[] outArray=new String[fieldCnt];
    //Next, go through and pick out each of the fields
    int currentLoc=0;
    int whichString=0;
    int i=0;
    boolean done=false;
    inQuotes=false;

    do {
      StringBuffer sb=new StringBuffer();
      char ch;
      boolean loopDone=false;
      do {
        ch=s.charAt(i);
        if (ch == '\"') inQuotes = !inQuotes;
        else if (inQuotes) sb.append(ch); //everything in quotes goes in
        else if (ch == sepBy) loopDone=true;
        else sb.append(ch);
        if (++i >= strLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichString++]=sb.toString();
    } while(!done);
    return outArray;
  }

  /**Splits the given string using the given separator char. Returns and array of Strings.
   * Starts at the specified offset location. 
   * Note: if offset==-1 then adjusts offset to automatically skip over a leading separator character. When splitting on RiriSeparators, the starting separator can be a problem (sometimes there and sometimes not)
   * Note: Does not return null
   * */
  public final static String[] toArrayIgnoreQuotes(String s, char sepBy, int offset) {
    int strLen= s==null ? 0 : s.length();
    if(offset<0) { //do auto-skip for leading riri separator
      if(strLen<1) return StringA0; //notshing to do
      if(s.charAt(0)==sepBy) offset=1;
      else offset=0;
    }
    if((strLen-offset) <= 0) return StringA0; //ie: String[0]

    //First, count through and allocate array:
    int fieldCnt=0;
    for(int i=offset; i<strLen; i++)
      if(s.charAt(i)==sepBy) fieldCnt++;
    fieldCnt++; //one more for good measure
    String[] outArray=new String[fieldCnt];

    //Go through it again, this time picking out each of the fields:
    int whichString=0;
    int i=offset;
    boolean done=false;
    do {
      StringBuffer sb=new StringBuffer();
      char ch;
      boolean loopDone=false;
      do {
        ch=s.charAt(i);
        if(ch==sepBy) loopDone=true;
        else sb.append(ch);
        if(++i >= strLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichString++]=sb.toString();
    } while(!done);
    return outArray;
  }

  /**Splits the given string using the given separator chars. Returns and array of Strings.
   * Starts at the specified offset location. 
   * Note: if offset==-1 then adjusts offset to automatically skip over a leading separator character. When splitting on RiriSeparators, the starting separator can be a problem (sometimes there and sometimes not)
   * Note: Does not return null
   * */
  public final static String[] toArrayIgnoreQuotes(String s, char sepBy1, char sepBy2, int offset) {
    int strLen= s==null ? 0 : s.length();
    if(offset<0) { //do auto-skip for leading riri separator
      if(strLen<1) return StringA0; //nothing to do
      if(s.charAt(0)==sepBy1) offset=1;
      else if(s.charAt(0)==sepBy2) offset=1;
      else offset=0;
    }
    if((strLen-offset) <= 0) return StringA0; //ie: String[0]

    //First, count through and allocate array:
    int fieldCnt=0;
    for(int i=offset; i<strLen; i++) {
      if(s.charAt(i)==sepBy1) fieldCnt++;
      else if(s.charAt(i)==sepBy2) fieldCnt++;
    }
    fieldCnt++; //one more for good measure
    String[] outArray=new String[fieldCnt];

    //Go through it again, this time picking out each of the fields:
    int whichString=0;
    int i=offset;
    boolean done=false;
    do {
      StringBuffer sb=new StringBuffer();
      char ch;
      boolean loopDone=false;
      do {
        ch=s.charAt(i);
        if(ch==sepBy1) loopDone=true;
        else if(ch==sepBy2) loopDone=true;
        else sb.append(ch);
        if(++i >= strLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichString++]=sb.toString();
    } while(!done);
    return outArray;
  }

  /**Similar as toArray except works with bytes instead of String
   * Also, does not check for quoted strings*/
  public final static byte[][] toByteArray(String s, char sepBy) {
    int strLen = s==null ? 0 : s.length();
    if (strLen == 0) return null;

    //First, Count the sepBy chars so can allocate array:
    int fieldCnt=0;
    int strLenM1 = strLen-1;
    //s = s.trim(); //needed?
    for (int i=0; i < strLenM1; i++) //Check all but last character (don't care if separator is the last character as in "011110;"
      if(s.charAt(i)==sepBy) fieldCnt++;

    if(s.charAt(strLenM1)==sepBy)
      --strLen; //Dump the final char if is separator

    if(fieldCnt==1) { //simplify single entry items (used heavily for vector entries, e.g., "101010001;" )
      byte[][] outArray = new byte[1][strLen];
      for(int i=0; i<strLen; i++)
        outArray[0][i]=(byte)s.charAt(i);
      return outArray;
    }

    if(s.charAt(strLenM1)==sepBy) {
      s = s.substring(0,strLenM1); //Dump the final char if is separator
      strLen = s.length();
      if(strLen==0) //special case where continuation character is all there is
        return new byte[1][0];
    }

    fieldCnt++; //One more for good measure
    byte[][] outArray=new byte[fieldCnt][];
    //Next, go through and pick out each of the fields
    int whichString=0;
    int i=0;
    boolean done=false;

    do {
      ByteBuffer bb=new ByteBuffer();
      char ch;
      boolean loopDone=false;
      do {
        ch=s.charAt(i);
        if(ch==sepBy) loopDone=true;
        else bb.append((byte)ch);
        if (++i >= strLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichString++]=bb.toBytes();
    } while(!done);
    return outArray;
  }

  /**Separates the given byte array into a set of byte arrays, splitting on any of the given list of sepBy characters
   * This version returns it as an array of Strings*/
  public final static String[] toStringArray(byte[] baIn, byte[] sepBy) {
    return toStringArray(baIn, -1, sepBy);
  }

  /**Separates the given byte array into a set of byte arrays, splitting on any of the given list of sepBy characters
   * This version returns it as an array of Strings
   * Allows specifying valid length of input array. -1 uses entire array*/
  public final static String[] toStringArray(byte[] baIn, int len, byte[] sepBy) {
    if(len<0) len = baIn==null ? 0 : baIn.length;
    byte[][] baa = toByteArray(baIn, len, sepBy);
    len = baa == null ? 0 : baa.length;
    if(len==0) return null;
    String[] sa = new String[len];
    for(int i=0; i<len; i++)
      sa[i] = new String(baa[i]);
    return sa;
  }

  /**Separates the given byte array into a set of byte arrays, splitting on any of the given list of sepBy characters*/
  public final static byte[][] toByteArray(byte[] baIn, byte[] sepBy) {
    return toByteArray(baIn, -1, sepBy);
  }

  /**Separates the given byte array into a set of byte arrays, splitting on any of the given list of sepBy characters
   * Allows specifying valid length of input array. -1 uses entire array*/
  public final static byte[][] toByteArray(byte[] baIn, int len, byte[] sepBy) {
    if (baIn == null) return null;
    int baInLen=baIn.length;
    if (baInLen == 0) return null;
    if( (len>=0) && (len < baInLen) ) baInLen=len; //see if need to reduce the number of items converted
    int sepByLen = sepBy==null ? 0 : sepBy.length;

    byte[][] outArray;
    if(sepByLen==0) { //No separators means send out what was received
      outArray = new byte[1][];
      outArray[0] = baIn;
      return outArray;
    }

    //First, Count the sepBy chars so can allocate array:
    int fieldCnt=0;

    for (int i=0; i < baInLen; i++) {
      for(int j=0; j<sepByLen; j++) {
        if(baIn[i]==sepBy[j]) fieldCnt++;
      }
    }
    fieldCnt++; //One more for good measure

    outArray=new byte[fieldCnt][];
    //Next, go through and pick out each of the fields
    int currentLoc=0;
    int whichArray=0;
    int i=0;
    boolean done=false;

    do {
      ByteBuffer bb=new ByteBuffer();
      char ch;
      boolean loopDone=false;
      do {
        byte b =baIn[i];
        for(int j=0; j<sepByLen; j++) {
          if(b == sepBy[j]) loopDone=true;
        }
        if(!loopDone) bb.append(b);
        if (++i >= baInLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichArray++]=bb.toBytes();
    } while(!done);
    return outArray;
  }

  /**Separates the given byte array into a set of byte arrays, splitting on the given sepBy byte*/
  public final static byte[][] toByteArray(byte[] baIn, byte sepBy) {
    int baInLen = (baIn==null) ? 0 : baIn.length;
    if (baInLen == 0) return null;

    //Count the sepBy chars so can allocate array:
    int fieldCnt=0;
    for(int i=0; i<baInLen; i++)
      if(baIn[i]==sepBy) fieldCnt++;
    fieldCnt++; //one more for good measure
    byte[][] outArray=new byte[fieldCnt][];

    //Pick out each of the fields:
    int currentLoc=0;
    int whichArray=0;
    int i=0;
    boolean done=false;
    do {
      ByteBuffer bb=new ByteBuffer();
      char ch;
      boolean loopDone=false;
      do {
        byte b = baIn[i];
        if(b == sepBy) loopDone=true;
        if(!loopDone) bb.append(b);
        if (++i >= baInLen) {
          loopDone=true;
          done=true;
        }
      } while(!loopDone);
      outArray[whichArray++]=bb.toBytes();
    } while(!done);
    return outArray;
  }

  /**General purpose, static, utility method to turn a String array into
   *      a Comma Separated String*/
  public final static String toCsv(String[] sa) {
    if (sa == null) return "";
    return makeCsv(sa, 0, sa.length - 1);
  }

  /**General purpose, static, utility method to turn a collection into
   * a Comma Separated String. The toString() method is called on each
   * entry encountered, unless it is itself a Collection.
   * If a collection is encountered, it is emitted at the next level
   * down of separator.
   * The separators used at each level is given by the 'separators' parameter.
   * By default (if given a null) it is set to the sequence: ", ; | &"
   * So that 1st level items are comma separated, second level items are semicolon separated etc.
   * If more levels are encountered than the available separators the following
   * syntax is used: item {n} item...
   * Where {n} is the syntax used to signify a separator at level 'n' (1 based)
   * TODO: Add Arrays (at minimum String[]) to recursion. Also modify toCsv(String[]) to same as this.
   * */
  public final static String toCsv(Collection col, String separators) {
    return toCsv(col, separators, 0);
  }

  /**Convenience method for situation where default separators are all that is needed*/
  public final static String toCsv(Collection col) {
    return toCsv(col, null, 0);
  }

  /**This is the internal version of toCsv(Collection), for recursion. The version
   * without the 'level' parameter is the user (public) version.
   * */
  private final static String toCsv(Collection col, String separators, int level) {
    if(col==null) return null; //Just in case
    int slen = separators == null ? 0 : separators.length();
    if(separators==null) separators=",;|&"; //Default Separators
    boolean needsSeparator = (level==0) ? false : true; //Used to suppress leading separator on very first, top level, item

    StringBuffer sb = new StringBuffer();
    Iterator iter = col.iterator();
    while(iter.hasNext()) {
      Object n = iter.next();
      if(n instanceof Collection)
        sb.append(toCsv( ((Collection)n), separators, level+1));
      else if(n!=null) {
        if(needsSeparator) //Suppresses separator the first time thru at the top level
          sb.append(getSeparatorForLevel(separators, level));
        else needsSeparator=true;
        sb.append(n.toString());
      }
      //else n==null which just goes in as an empty, delimited, entry.
    }
    return sb.toString();
  }

  private final static String getSeparatorForLevel(String separators, int level) {
    int len = separators ==null ? 0 : separators.length();
    if(level<len) return Character.toString(separators.charAt(level));
    return " {"+(level+1)+"} "; //If number of levels encountered exceeds number of available separators
  }

  /**General purpose, static, utility method to turn a String array into
   *      a Comma Separated String.  Converts starting at given (zero based) index*
   */
  public final static String toCsv(String[] sa, int startIndex) {
    if (sa == null) return "";
    return makeCsv(sa, startIndex, sa.length - 1);
  }

  /**Separates a CSV String into an array of doubles*/
  public final static double[] toDoubleArray(String sInput) {
    String[] sa=toArray(sInput);
    int len=(sa == null) ? 0 : sa.length;
    double[] da=new double[len];
    for (int i=0; i < len; i++)
      da[i]=(sa[i] == null) ? 0.0 : Double.parseDouble(sa[i].trim());
    return da;
  }

  /**Separates a CSV String into an array of int*/
  public final static int[] toIntArray(String sInput) {
    String[] sa=toArray(sInput);
    int len=(sa == null) ? 0 : sa.length;
    int[] ia=new int[len];
    for (int i=0; i < len; i++)
      ia[i]=(sa[i] == null) ? 0 : Integer.parseInt(sa[i].trim());
    return ia;
  }

}
