package ri.util;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;

public class DbaseUtil {

  /**Private Constr*/
  private DbaseUtil() {}
  
  /**Utility function to connect to an SQLite database
   * Throws ClassCastException if the JDBC driver class is not available
   * Throws SQLException if fails in establishing a connection to the specified database
   * */
  public static Connection doConnectToSqlite(String dbName, boolean verbose) throws SQLException, ClassNotFoundException {
    dbName = (dbName==null) ? "" : dbName.trim();

    String jdbcClassName = "org.sqlite.JDBC";
    Class.forName(jdbcClassName); //ensure the sqlite JDBC class is available (throws ClassNotFoundException) 
    String connectTo = "jdbc:sqlite:"+dbName;
    if(verbose)
      Emit.out("Connect using: '"+connectTo+"' (driver="+jdbcClassName+")");
    return DriverManager.getConnection(connectTo); //throws SQLException
  }
  
  
  /**Utility function to connect to a database
   * Parameters:
   *     jdbcType: either a short name (e.g. postgresql, db2, sqlite, hsqldb) or else 
   *               a fully qualfied name for the JDBC driver class (e.g. org.postgresql.Driver)
   *     ip: optional parameter (null is ok) to specify the database connection
   *     isNonSqliteFile: set to true for a file based dbase (i.e. "file:" gets prepended to the given dbName). 
   *                      Note: set this to false for sqlite (even though it is a file)
   *     dbName: the name of the database to be connected to (for a file based database this is the filename)
   *     login: user name for the database login (null is ok) 
   *     pwd: user password for the database login (null is ok)
   *     verbose: if this is true then 'Emit' statements are generated indicating the connection status.
   *      
   * Throws:
   *    ClassCastException: if the JDBC driver class is not available
   *    SQLException: if fails in establishing a connection to the specified database
   * */
  public static Connection doConnectTo(String jdbcType, String ip, boolean isNonSqliteFile, String dbName, String login, String pwd, boolean verbose) throws SQLException, ClassNotFoundException {
    //trim allows using " " (space) for values (which allows using that to specify an empty entry in a PComboBox):    
    jdbcType = (jdbcType==null) ? "" : jdbcType.trim();
    ip       = (ip==null)       ? "" : ip.trim();
    dbName   = (dbName==null)   ? "" : dbName.trim();
    login    = (login==null)    ? "" : login.trim();
    pwd      = (pwd==null)      ? "" : pwd.trim();
    
    //Load JDBC-ODBC bridge:
    String jdbcClassName=null;
    if(jdbcType.equalsIgnoreCase("postgresql"))  jdbcClassName = "org.postgresql.Driver"; 
    else if(jdbcType.equalsIgnoreCase("db2"))    jdbcClassName = "COM.ibm.db2.jdbc.app.DB2Driver";
    else if(jdbcType.equalsIgnoreCase("sqlite")) jdbcClassName = "org.sqlite.JDBC";
    else if(jdbcType.equalsIgnoreCase("hsqldb")) jdbcClassName = "org.hsqldb.jdbcDriver";
    else jdbcClassName = jdbcType; //assume is a fully qualified class name if is not one of the known abbreviations

    if(jdbcClassName!=null)
      Class.forName(jdbcClassName); //throws ClassNotFoundException 
    else if(verbose) 
      Emit.out("Warning. doConnectTo(): Unknown JDBC Type: ["+jdbcType+"] Unable to load Driver.");

    StringBuffer connectTo = new StringBuffer(); // "jdbc:postgresql://localhost/DB_NAME"
    connectTo.append("jdbc:");
    connectTo.append(jdbcType);
    connectTo.append(":");
    if(ip!=null && ip.length()>0) {
      connectTo.append("//");
      connectTo.append(ip);
      connectTo.append('/');
    }
    else if(isNonSqliteFile) //note: this one is set to false for Sqlite (even though is a file)
      connectTo.append("file:");
    connectTo.append(dbName);
    
    boolean hasLogin = (login!=null && login.length()>0);
    boolean hasPwd = (pwd!=null && pwd.length()>0);

    if(verbose)
      Emit.out("Connect using: '"+connectTo+"' (driver="+jdbcClassName+")");
    if(hasLogin || hasPwd) 
      return DriverManager.getConnection(connectTo.toString(), login, pwd); //throws SQLException
    else
      return DriverManager.getConnection(connectTo.toString()); //throws SQLException
  }
  
  
  /**Debug utility to display the results of the given query*/
  public static void showResultsOfQuery(String msg, Statement sqlStatement, String sqlQuery) {
    Emit.out("--------------------------------");
    if(msg!=null) Emit.out(msg);
    Emit.out("Query: ["+sqlQuery+"]:");

    ResultSet resultSet=null;
    try {
      resultSet=sqlStatement.executeQuery(sqlQuery);
      ResultSetMetaData rsmd = resultSet.getMetaData();
      int colCnt = rsmd.getColumnCount();
      Emit.outNoCr("-->"+colCnt+" columns:");
      for(int i=1; i<=colCnt; i++)
        Emit.outNoCr("\t"+rsmd.getColumnName(i));
      Emit.out("");
      while(resultSet.next()) {
        for(int i=1; i<=colCnt; i++) {
          Emit.outNoCr("\t"+resultSet.getString(i));
        }
        Emit.out("");
      }
      Emit.out("--------------------------------");
    }
    catch (SQLException e) {
      Emit.out("Error in "+DbaseUtil.class.getName()+".showResultsOfQuery(): Unable to perform query ["+sqlQuery+"] because: "+e);
      e.printStackTrace();
    }
    finally {
      try { resultSet.close(); } catch(Exception e) {}
    }
  }
  
  
  /**Debug utility to display the results of the given query*/
  public static void showResultsOfQuery(String msg, Connection sqlConnection, String sqlQuery) {
    try { showResultsOfQuery(msg, sqlConnection.createStatement(), sqlQuery); }
    catch (SQLException e) {
      Emit.out("Error in "+DbaseUtil.class.getName()+".showResultsOfQuery(): Unable to create a dbase 'statement' (for query ["+sqlQuery+"]): "+e);
    }
  }
  
}
