package ri.util;

import java.awt.Color;
import java.awt.Component;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingUtilities;


/**General purpose emit display. 
 * Use this instead of System.out to feed text messages to an appropriate display panel.
 * 
 * Text areas 'out' and 'err' are provided via the accessor methods: checkoutTextArea() and checkoutErrTextArea().
 * Output is sent to the messaging areas via the static methods: Emit.out(...) and Emit.err(...).
 * The actual Emit component is a JScrollPane containing a JTextField that can be placed in an application as required.
 
 * out(), outNoCr(), err() and errNoCr() send text to the relevant display area
 * checkinTextArea(), checkinErrTextArea(): provides a way to give back the Emit component and releases it from being the emit window.
 * clear(), clearErr(): wipes the display area clean. 
 * setBackgroundColor(): sets the text area background color 
 * updateDisplay(): will do a printImmediately 
 * setLineWrap(): turns line wrap mode on and off. If off then long lines extend to the right without wrap 
 * setWrapStyleWord(): optional wrap at word boundaries 
 * setTextAreaCols(): Sets a preferred number of columns for display (useful for BorderLayout.EAST/WEST) 
 * setTextAreaRows(): Sets a preferred number of rows for display (useful for BorderLayout.NORTH/SOUTH) 
 * enableTimestamps(): optional emit of timestamp for each displayed entry 
 * setMouseListener(): allows replacing default mouse listener with a user supplied one
 * addMouseListener(): adds a mouse listener to the text area
 * enableTimestamps(boolean): enables/disables the displaying of timestamps with each entry
 * captureConsoleOut(boolean): turns on and off the capturing of stdOut/stdErr to the Emit window
 *
 *  NOTE: When checkOutTextArea is called, the error text area is assigned to the same (if errTxtArea==null)
 *
 *TODO: Restore the err text area
 *
 *History:
 *  3/7/2013 OER: Added line wrap toggle to the default context (right click) menu 
 *  4/5/2012 OER: Added captureConsoleOut() to allow Emit to capture stdOut and stdErr
 *  1/28/12 OER: added 'synchronized' to checkoutTextArea() to prevent a race condition on multiple opens in quick succession.
 *  12/13/09 OER: 
 *    -Removed errTextArea and altTextArea. err() now just always writes to text area
 *  12/1/06 OER: 
 *    -added timestamp option, added 'final' to all methods
 *  9/16/1999 OER:
 *    -Updated to package ri.oer.rix.panel
 *  3/1998 OER: created.                 
 */
public class Emit {

  private static JTextArea _tarea = null; //main output text area
  private static Color _bgndColor; //background color
  private static JScrollPane _scrollpane = null;
  private static boolean _timestampsEnabled = false;
  private static boolean _useDefaultMouseListener = true;
  private static SimpleDateFormat _timestampFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

  //These are used to control the amount of content held by _tarea at any one time:
  private static int _idealSize = 100000;  //after a cleanup, this is how much _tarea will be holding
  private static int _maxExcess = 200000;  //how much excess bytes (over the 'ideal') before _tarea gets 'cleaned up'

  private static boolean _captureConsoleOut = false; //if true then stdOut and stdErr are going to the Emit window
  private static PrintStream _savOut=null; //used to preserve the original console stdout
  private static PrintStream _savErr=null; //used to preserve the original console stderr

  /**Used to clear the text area synchronously with Swing event thread*/
  private static Runnable clearIt = new Runnable() {
    public void run() {
      _tarea.setText("");
    }
  };

  /**Private Constr. No need to instantiate this class*/
  private Emit() {}

  /**Send this message when the GUI panel is done with the text area. Does does not dispose of or
   * modify the text area in any way. This class simply stops using it*/
  synchronized public static final void checkinTextArea() {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        _tarea = null;
      }
    });
  }

  /**Create and initialize a text area that can be placed in a panel for user I/O. Note only one
   * _tarea can be assigned at a time so use checkinTextArea in order to allow another panel to use.
   * Added synchronized 1/28/11: was having issue when multiple Whiteboard apps were opened in quick succession 
   * and only the first was supposed to have the 'Emit'*/
  synchronized public static final Component checkoutTextArea(boolean useDefaultMouseListener) {
    if(_tarea != null) return null;
    _useDefaultMouseListener = useDefaultMouseListener;
    if(SwingUtilities.isEventDispatchThread()) return doTheChkoutTxtArea();
    try {
      SwingUtilities.invokeAndWait(new Runnable() {
        public void run() { doTheChkoutTxtArea(); }
      });
      return _scrollpane;
    }
    catch(Exception e) { return null; }
  }
  
  /**Calling this causes stdOut and stdErr to be re-directed to the Emit window*/
  public synchronized static void captureConsoleOut(boolean b) {
    if(_captureConsoleOut==b) return; //already doing as requested

    _captureConsoleOut = b;      
    if(_captureConsoleOut) {
      OutputStream out = new OutputStream() {  
        public void write(int b) throws IOException { outNoCr(String.valueOf((char) b)); }  
        public void write(byte[] b, int off, int len) throws IOException { outNoCr(new String(b, off, len)); }  
        public void write(byte[] b) throws IOException { write(b, 0, b.length); }
      };  
      _savOut = System.out; //save the original values
      _savErr = System.err;
      System.setOut(new PrintStream(out, true));  
      System.setErr(new PrintStream(out, true));  
    }
    else { //stop capturing std out
      if(_savOut!=null) {  //restore the original stdout value
        System.setOut(_savOut);
        _savOut=null;
      }
      if(_savErr!=null) {  //restore the original stderr value
        System.setErr(_savErr);  
        _savErr=null;
      }
    }
  }

  /**Clears the display of the currently active text area*/
  public static final void clear() {
    if(_tarea != null) SwingUtilities.invokeLater(clearIt);
  }

  public static final void setLineWrap(boolean b) {
    _tarea.setLineWrap(b);
  }

  public static final void setWrapStyleWord(boolean b) {
    _tarea.setWrapStyleWord(b);
  }

  /**Outputs the given Exception (with a carriage return) to the currently active text area. Sends
   * it to stdout, if all else fails*/
  public static final void out(Exception e) {
    outNoCr(e.toString() + "\n");
  }

  /**Outputs the given Exception (with a carriage return) to the error text area. Sends it to
   * stdout, if all else fails*/
  public static final void err(Exception e) {
    outNoCr(e.toString() + "\n");
  }

  /**Outputs the given String (with a carriage return) to the currently active text area. Sends it
   * to stdout, if all else fails*/
  public static final void out(String s) {
    outNoCr(s + "\n");
  }

  /**Outputs the given String (with a carriage return) to the error text area. Sends it to stdout,
   * if all else fails*/
  public static final void err(String s) {
    outNoCr(s + "\n");
  }

  /**Outputs the given String (w/o carriage return) to the currently active text area. Sends it to
   * stdout, if all else fails*/
  public static final void outNoCr(String s) {
    if(_timestampsEnabled) s = getTimestamp() + " " + s;
    final String outMsg = s;
    if(_tarea != null) {
      SwingUtilities.invokeLater(new Runnable() { //if for no other reason, this is required so that EmitDocListener doesn't cause a hang when it auto-scrolls
        public void run() {
          _tarea.append(outMsg);
          
          //keep the text area down to a certain character size:
          int excess = _tarea.getDocument().getLength() - _idealSize;
          if (excess >= _maxExcess)
            _tarea.replaceRange("", 0, excess);
        }
      });
    }
    else {
      if(!_captureConsoleOut) //prevents recursion
        System.out.print(s);
    }
  }

  /**Accessor for the actual (internal) text area. for use by external event handlers and such. Call
   * this only after checkoutTextArea() to ensure all is properly set up*/
  public static final JTextArea getTextArea() {
    return _tarea;
  }

  /**Sets the background color of the text area*/
  public static final void setBackgroundColor(Color color) {
    _bgndColor = color;
    if(_tarea != null) {
      SwingUtilities.invokeLater(new Runnable() {
        public void run() {
          _tarea.setBackground(_bgndColor);
        }
      });
    }
  }

  /**Sets the number of columns in the text area*/
  public static final void setTextAreaCols(int colCnt) {
    if(_tarea != null) _tarea.setColumns(colCnt);
  }

  /**Sets the number of rows in the text area*/
  public static final void setTextAreaRows(int rowCnt) {
    if(_tarea != null) _tarea.setRows(rowCnt);
  }

  /**Update the display area right away*/
  public static final void updateDisplay() {
    if(_tarea != null) _tarea.paintImmediately(new Rectangle(_tarea.getSize()));
  }

  /**If set to true, a timestamp precedes each emitted string Default is timestamps=false*/
  public static final void enableTimestamps(boolean b) {
    _timestampsEnabled = b;
  }

  /**Sets the output text area up with the given mouse listener Note: to use this externally,
   * checkoutTextArea(false) is recommended so that two handlers are not present.*/
  public static final void addMouseListener(MouseListener ml) {
    if(_tarea!=null)
      _tarea.addMouseListener(ml);
  }

  /**Provides a default mouse listener for use by the output text area. This listener provides right
   * mouse click options to perform: Clear, Copy, Paste, Timestamp toggle, Linewrap toggle*/
  public static final MouseListener getDefaultMouseListener() {
    MouseListener ml = new MouseListener() {
      public void mouseClicked(MouseEvent e) {
        int mod = e.getModifiers();
        if((mod & MouseEvent.BUTTON3_MASK) != 0) {
          ActionListListener actn = new ActionListListener() {
            String[] _actions = new String[] { "Clear", "Copy", "Paste", "Timestamps", "Line wrap" };

            public Object[] getActions() { return _actions; }

            public void doAction(int index, MouseEvent e) {
              String s;
              switch(index) {
                case 0: //Clear
                  _tarea.setText("");
                  break;
                case 1: //Copy
                  s = _tarea.getSelectedText(); //get selected
                  if(s == null || s.length() == 0) s = _tarea.getText(); //get it all
                  ClipboardUtil.putToClipboard(s);
                  break;
                case 2: //Paste
                  s = ClipboardUtil.getClipboardAsString(null);
                  if(s != null) _tarea.insert(s, _tarea.getCaretPosition());
                  break;
                case 3: //toggle timestamps
                  _timestampsEnabled = !_timestampsEnabled;
                  out("Timestamps "+(_timestampsEnabled ? "Enabled" : "Disabled"));
                  break;
                case 4: //toggle line wrap
                  boolean doLineWrap = !_tarea.getLineWrap();
                  _tarea.setLineWrap(doLineWrap);
                  _tarea.setWrapStyleWord(true);
                  out("Line wrap "+(doLineWrap ? "Enabled" : "Disabled"));
                  break;
              }
            }
          };
          new ActionList(_tarea, e, actn).setVisible(true);
        }
      }
      public void mouseEntered(MouseEvent e) {}
      public void mouseExited(MouseEvent e) {}
      public void mousePressed(MouseEvent e) {}
      public void mouseReleased(MouseEvent e) {}
    };
    return ml;
  }

  // -----------------------------
  // Private methods:
  // -----------------------------

  /**Factory method for JTextAreas suitable for our little emits*/
  private static final JTextArea createTextArea() {
    JTextArea tarea = new JTextArea();
    tarea.setLineWrap(true);
    tarea.setWrapStyleWord(true);
    tarea.setAutoscrolls(false);
    return tarea;
  }

  /**Used by checkoutTextArea*/
  private static final Component doTheChkoutTxtArea() {
    _tarea = createTextArea();
    _tarea.setLineWrap(false);
    _scrollpane = new JScrollPane(_tarea);
    _scrollpane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
    _tarea.getDocument().addDocumentListener(new EmitDocListener(_scrollpane)); //causes any newly appended text to be scrolled into view
    if(_useDefaultMouseListener) addMouseListener(getDefaultMouseListener());
    return _scrollpane;
  }

  /**Creates and returns a formatted date/time string*/
  private static final String getTimestamp() {
    return _timestampFormat.format(new Date());
  }
}
