package ri.util;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.prefs.Preferences;

import javax.swing.MutableComboBoxModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

/**A ComboBoxModel with Persistence.
 * Remembers entries made from previous invocations
 * 
 * This version uses the Java Preferences API (http://docs.oracle.com/javase/7/docs/technotes/guides/preferences/)
 *  
 * Creation date: derived from PersistentComboBoxModel on 1/10/2014
 * @author: OER
 */
public class PComboBox2_Model implements MutableComboBoxModel {

  private Class _clazz;                    //class for the item being persisted
  private String _name;                    //name for this object (for persistent store)
  private int _maxEntries;                 //maximum # of entries this model will persist
  private List _entries=new ArrayList();   //the list of values held by this combo box model
  private String _default;                 //default value required for Java Preferences API. Note: this is the value used in case the backing store is not available

  private List _listeners;                 //listeners for events generated by this model
  private Object _selectedItem=null;       //this represents the currently selected item
  private final ListDataEvent listDataEvent_AddedAt0 = new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED, 0, 0); //for reuse
  

  /**Constr. Given values for:
   *   clazz: Java class of object being persisted (can be null). Note: if null then the class of this box model is used  
   *   name: unique name (within this class) for the value being persisted
   *   maxEntries: maximum for number of entries to persist (Note: -1 indicates no limit)
   *   presets: Initial seed values. Note: the Java Preferences API requires there be at least 1 value in this list (to use as the default)
   * */
  public PComboBox2_Model(Class clazz, String name, int maxEntries, String[] presets) {
    _clazz=clazz;
    _name=name;
    _maxEntries=maxEntries;
    initFromPersistentStore();
    int len = (presets==null) ? 0 : presets.length;
    for(int i=0; i < len; i++)
      addItemIfDnExist(presets[i], false, false);
    _default = len>0 ? presets[0] : "";
    updatePersistentStore();
    emitContentsChanged(0, getSize()); //update all listeners
  }


  /**Constr. Given values for:
   *   clazz: Java class of object being persisted (can be null). Note: if null then the class of this box model is used  
   *   name: unique name (within this class) for the value being persisted
   *   maxEntries: maximum for number of entries to persist (Note: -1 indicates no limit)
   *   csvPresets: comma separated string representing the initial seed values to persist. 
   *               Note: the Java Preferences API requires there be at least 1 value in this list (to use as the default)
   * */
  public PComboBox2_Model(Class clazz, String name, int maxEntries, String csvPresets) {
    this(clazz, name, maxEntries, Csv.toArray(csvPresets));
  }

  /**Adds an item to the end of the model
   * Note: Required for interface MutableComboBoxModel*/
  public void addElement(Object ob) {
    addItem(ob, false, true);
  }

  /**Adds to list of items held by this object, at the specified index location
   * (Note, this does not update persistent store)
   * If item is already in list, places it at the beginning.
   * If emitUpdate is true then listeners are updated*/
  public void addItem(Object ob, int index, boolean emitUpdate) {
    if(ob == null) return;
    String s=ob.toString();
    if(s.length() == 0) return;
    _entries.remove(s);     //remove it if it already exists
    _entries.add(index, s); //place new entry at specified location

    if(emitUpdate) 
      emitContentsChanged(0, getSize()); //JComboBox seems to require this
    if(getSize() == 1 && _selectedItem == null) 
      setSelectedItem(ob); //JComboBox seems to require this

    boolean trimmed=trimAsNeeded(); //trim off excess items as required
    if(trimmed && emitUpdate) 
        emitContentsChanged(0, getSize());
  }

  /**Adds to list of items held by this object (Note, this does not update
   * persistent store.  If item is already in list, places it at the beginning.
   * If atBeginning is true then new items are inserted at start of list
   * If emitUpdate is true then listener are updated*/
  public void addItem(Object ob, boolean atBeginning, boolean emitUpdate) {
    if(ob == null) return;
    String s=ob.toString();
    if(s.length() == 0) return;
    _entries.remove(s);                  //remove it if it already exists
    if(atBeginning) _entries.add(0, s); //place new entries at the front
    else _entries.add(s);                //place new entries at end

    if(emitUpdate) 
      emitContentsChanged(0, getSize()); //JComboBox seems to require this
    if(getSize() == 1 && _selectedItem == null) 
      setSelectedItem(ob); //JComboBox seems to require this

    boolean trimmed=trimAsNeeded();      //trim off excess items as required
    if(trimmed && emitUpdate) 
      emitContentsChanged(0, getSize());
  }

  /**Similar to addItem() except no modification of list ordering occurs.
   * If item already exists in the list, it is left unchanged*/
  public void addItemIfDnExist(Object ob, boolean atBeginning, boolean emitUpdate) {
    if(ob == null) return;
    String s=ob.toString();
    if(s.length() == 0) return;
    if(_entries.contains(s)) return; //if already in list then do nothing

    if(atBeginning) {
      _entries.add(0, s); //place new entries at the front
      if(emitUpdate) emitIntervalAdded(0, 0); //all we did was add at loc 0
    }
    else {
      _entries.add(s); //place new entries at end
      int loc=getSize() - 1;
      if(emitUpdate) emitIntervalAdded(loc, loc);
    }

    boolean trimmed=trimAsNeeded(); //trim off excess items as required
    if(trimmed && emitUpdate) 
      emitContentsChanged(0, getSize());
  }

  /** Add a listener to the list that's notified each time a change
   * to the data model occurs.
   * Note: Required for interface ListModel (part of ComboBoxModel)*/
  public void addListDataListener(ListDataListener l) {
    if(_listeners == null) _listeners=new ArrayList();
    _listeners.add(l);
  }

  /**Just updates the whole list for simplicity (rather than sorting through all the individual changes*/
  public void emitContentsCompletelyChanged() {
    emitContentsChanged(0, getSize());
  }

  /**Emit changes to model to all ListDataListener objects.
   * Note: Interval includes both indices*/
  private void emitContentsChanged(int ndx0, int ndx1) {
    if(_listeners == null) return;
    ListDataEvent lde=new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, ndx0, ndx1);
    int len=_listeners.size();
    for(int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).contentsChanged(lde);
  }

  /**Emit changes to model to all ListDataListener objects
   * Note: Interval includes both indices*/
  private void emitIntervalAdded(int ndx0, int ndx1) {
    if(_listeners == null) return;

    ListDataEvent lde;
    if(ndx0 == 0 && ndx1 == 0) lde=listDataEvent_AddedAt0; //usu. items are added at loc. 0
    else lde=new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED, ndx0, ndx1);

    int len=_listeners.size();
    for(int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).intervalAdded(lde);
  }

  /**Emit changes to model to all ListDataListener objects
   * Note: Interval includes both indices*/
  private void emitIntervalRemoved(int ndx0, int ndx1) {
    if(_listeners == null) return;
    ListDataEvent lde=new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED, ndx0, ndx1);
    int len=_listeners.size();
    for(int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).intervalRemoved(lde);
  }

  /**Returns the value at the specified index.
   * Note: Required for interface ListModel (part of ComboBoxModel)*/
  public Object getElementAt(int ndx) {
    try { return _entries.get(ndx); }
    catch (Exception e) { return null; }
  }

  /**Return the selected item
   * Note: Required for interface ComboBoxModel (part of MutableComboBoxModel)*/
  public Object getSelectedItem() {
    return _selectedItem;
  }

  /**Returns the length of the list.
   * Note: Required for interface ListModel (part of ComboBoxModel)*/
  public int getSize() {
    return _entries.size();
  }

  /**Adds an item at a specific index
   * Note: Required for interface MutableComboBoxModel*/
  public void insertElementAt(Object ob, int index) {
    addItem(ob, index, true);
  }

  /**Remove all items held in the combo box list
   * Emits update to listeners*/
  public void removeAllItems(boolean emitUpdate) {
    int len=getSize();
    if(len == 0) return; //nothing to do
    for(int i=0; i < len; i++)
      _entries.remove(0);
    setSelectedItem(null);
    if(emitUpdate) emitIntervalRemoved(0, len); //inform the listeners
  }

  /**Removes an item from the model
   * Note: Required for interface MutableComboBoxModel*/
  public void removeElement(Object ob) {
    removeItem(ob, true);
  }

  /**Removes an item at a specific index
   * Note: Required for interface MutableComboBoxModel*/
  public void removeElementAt(int index) {
    removeItem(index, true);
  }

  /**Remove specified item held in the combo box list
   * Emits update to listeners if emitUpdate is true*/
  public void removeItem(int which, boolean emitUpdate) {
    int len=getSize();
    if(which >= len) return; //nothing to do

    _entries.remove(which);
    if(which == 0) 
      setSelectedItem(getSize() <= 1 ? null : getElementAt(which + 1)); //JComboBox seems to require this:
    else 
      setSelectedItem(getElementAt(which - 1));
    if(emitUpdate) 
      emitIntervalRemoved(which, which); //inform the listeners
  }

  /**Remove specified item held in the combo box list
   * Emits update to listeners if emitUpdate is true*/
  public void removeItem(Object ob, boolean emitUpdate) {
    _entries.remove(ob);
    if(emitUpdate) 
      emitContentsChanged(0, getSize()); //inform the listeners
  }

  /**Remove a listener from the list that's notified each time a
   * change to the data model occurs.
   * Note: Required for interface ListModel (part of ComboBoxModel) */
  public void removeListDataListener(ListDataListener l) {
    if(_listeners == null) return;
    _listeners.remove(l);
  }

  /**Set the selected item
   * Note: Required for interface ComboBoxModel (part of MutableComboBoxModel)*/
  public void setSelectedItem(Object anItem) {
    _selectedItem=anItem;
    addItem(_selectedItem, true, true);
    updatePersistentStore();
  }

  /**Trims the number of items held by this object
   * (as specified by 'maxEntries')
   * Returns true if any items were trimmed off*/
  private boolean trimAsNeeded() {
    //See if need to trim the list:
    if(_maxEntries < 0) return false; //no trimming for max < 0
    int currentSize=getSize();
    boolean trimmed=false;
    while(currentSize > _maxEntries) {
      _entries.remove(currentSize - 1);
      trimmed=true;
      currentSize=getSize();
    }
    return trimmed;
  }

  /**Initializes this object from file or guru or ?*/
  private void initFromPersistentStore() {
    if(_name == null) return; //needs a name in order for persistence to do its thing
    if(_clazz==null) _clazz=this.getClass(); //just in case, if no class is given then just use this object
    
    checkForOldPersistenceFile(true);    
    try {
      Preferences prefs = Preferences.userNodeForPackage(_clazz);
      String s = prefs.get(_name, _default);
      String[] sa = Csv.toArray(s);
      for(int i=0; i<sa.length; i++) {
        if(sa[i]==null) sa[i]="";
        addItem(sa[i], false, false);
      }
    }
    catch(Exception e) { System.out.println("Error reading in combo box persistence info for(" + _clazz.getName()+":"+_name + "): " + e); }
  }
  
  /**Initializes this object from an old style persistence file.
   * Note: does not call updatePersistentStore. Assumes this is done after.*/
  private void checkForOldPersistenceFile(boolean deleteExisting) {
    if (_name == null) return;
    BufferedReader in=null;
    File f=null;
    try {
      f = new File(_name + ".persistentCbox");
      if(!f.exists()) 
        return; //nothing to do

      in=new BufferedReader(new FileReader(f));
      while(in.ready()) {
        String s = in.readLine();
        if(s==null) break;
        addItem(s, false, false);
      }
    }
    catch (FileNotFoundException e) {} //ignore
    catch (Exception e) {
      System.out.println("Error reading in combo box persistence info for (" + _name + "): " + e);
    }
    finally { 
      if(in!=null) {
        try { in.close(); } catch (Exception e) {} 
      }
      if(f!=null && deleteExisting)
        f.delete(); //remove the existing persistence file
    }
  }

  

  /**Updates persistent store with the current state of this object
   * This version uses a file w/same name as object in the default directory*/
  private void updatePersistentStore() {
    if(_name == null) return; //no name == no persistence
    if(_clazz==null) _clazz=this.getClass(); //just in case, if no class is given then just use this object

    ArrayList list = new ArrayList();
    int len=getSize();
    String[] sa = new String[len];
    for(int i=0; i < len; i++) {
      Object ob=getElementAt(i);
      String s = ob==null ? "" : ob.toString();
      sa[i] = s;
    }
    String csv = Csv.toCsv(sa);

    try {
      Preferences prefs = Preferences.userNodeForPackage(_clazz);
      prefs.put(_name, csv);
    }
    catch(Exception e) {
      System.out.println("Error emitting combo box persistence info for("+_clazz.getName()+":"+_name + "): " + e);
    }
  }

  /**Returns an xml string representing the persistent content for the given item.
   * If null then the entire user settings tree is returned.
   * If user=true then user settings are returned, otherwise system settings.
   * */
  public static String getPersistenceTree(boolean user, Class clazz) {

    try {
      Preferences prefs;
      if(user) { //user settings
        if(clazz==null)
          prefs = Preferences.userRoot();
        else
          prefs = Preferences.userNodeForPackage(clazz);
      }
      else { //i.e. system settings
        if(clazz==null)
          prefs = Preferences.systemRoot();
        else
          prefs = Preferences.systemNodeForPackage(clazz);
      }
      
      //System.out.println("Prefs path="+prefs.absolutePath()); //debug
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      prefs.exportSubtree(baos);
      return baos.toString();
    }
    catch(Exception e) {
      Emit.out("Error in PComboBoxMode2.getPersistenceTree(): "+e);
      return null;
    }
  }
  
}
