package ri.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.MutableComboBoxModel;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;


/**A ComboBoxModel with Persistence.
 * Currently this uses a file in the default directory (named after the name given
 * when this object is created)
 * To do: Update this to be able to use guru instead of the file system.
 * Remembers entries made into it from previous invocations
 * Creation date: (5/28/02 12:25:24 PM)
 * @author: OER
 */
public class PersistentComboBoxModel implements MutableComboBoxModel {
  //public class PersistentComboBoxModel extends DefaultComboBoxModel {

  private static final String _ext=".persistentCbox"; //Filename extension for persistent store
  private String _name; //Name for this object (for persistent store. If null then no persistence will occur
  private int _maxEntries; //Maximum # of entries this model will persist
  private List _entries=new ArrayList(); //The list of values held by this combo box model
  //private List _entries = Collections.synchronizedList(new ArrayList());   //The list of values held by this combo box model

  private List _listeners; //Listeners for events generated by this model
  private Object _selectedItem=null; //This represents the currently selected item
  private ListDataEvent LDE_ADDED00=new ListDataEvent(this,
      ListDataEvent.INTERVAL_ADDED, 0, 0); //For reuse

  /**Constr: Given a name from which to instantiate,
   and a maximum for no of entries to persist
   (Note: max=-1 implies no limit)
   */
  public PersistentComboBoxModel(String name, int maxEntries) {
    this(name, maxEntries, (String[])null);
  }

  /**Constr: Given a name from which to instantiate,
   also given a maximum for no of entries to persist
   (Note: max=-1 implies no limit)
   And given some default values to include in the list
   */
  public PersistentComboBoxModel(String name, int maxEntries, String[] defaultVals) {
    _name=name;
    _maxEntries=maxEntries;
    initFromPersistentStore();
    if (defaultVals == null) return;
    int len=defaultVals.length;
    for (int i=0; i < len; i++)
      addItemIfDnExist(defaultVals[i], false, false);
    updatePersistentStore();
    emitContentsChanged(0, getSize()); //Update all listeners
  }


  /**Constr: Given a name from which to instantiate,
   also given a maximum for no of entries to persist
   (Note: max=-1 implies no limit)
   And a CSV string of default values to include in the list
   */
  public PersistentComboBoxModel(String name, int maxEntries, String csvPresets) {
    this(name, maxEntries, Csv.toArray(csvPresets));
  }

  /**Adds an item to the end of the model
   Note: Required for interface MutableComboBoxModel*/
  public void addElement(Object ob) {
    addItem(ob, false, true);
  }

  /**Adds to list of items held by this object, at the specified index location
   (Note, this does not update persistent store)
   If item is already in list, places it at the beginning.
   If emitUpdate is true then listeners are updated
   */
  public void addItem(Object ob, int index, boolean emitUpdate) {
    if (ob == null) return;
    String s=ob.toString();
    if (s.length() == 0) return;
    _entries.remove(s); //Remove it if it already exists
    _entries.add(index, s); //Place new entry at specified location

    if (emitUpdate) emitContentsChanged(0, getSize()); //JComboBox seems to require this
    if (getSize() == 1 && _selectedItem == null) setSelectedItem(ob); //JComboBox seems to require this

    boolean trimmed=trimAsNeeded(); //Trim off excess items as required
    if (trimmed) if (emitUpdate) emitContentsChanged(0, getSize());
  }

  /**Adds to list of items held by this object (Note, this does not update
   persistent store.  If item is already in list, places it at the beginning.
   If atBeginning is true then new items are inserted at start of list
   If emitUpdate is true then listener are updated
   */
  public void addItem(Object ob, boolean atBeginning, boolean emitUpdate) {
    if (ob == null) return;
    String s=ob.toString();
    if (s.length() == 0) return;
    _entries.remove(s); //Remove it if it already exists
    if (atBeginning) _entries.add(0, s); //Place new entries at the front
    else _entries.add(s); //Place new entries at end

    if (emitUpdate) emitContentsChanged(0, getSize()); //JComboBox seems to require this
    if (getSize() == 1 && _selectedItem == null) setSelectedItem(ob); //JComboBox seems to require this

    boolean trimmed=trimAsNeeded(); //Trim off excess items as required
    if (trimmed) if (emitUpdate) emitContentsChanged(0, getSize());
  }

  /**Similar to addItem() except no modification of list ordering occurs.
   If item already exists in the list, it is left unchanged.
   */
  public void addItemIfDnExist(Object ob, boolean atBeginning,
      boolean emitUpdate) {
    if (ob == null) return;
    String s=ob.toString();
    if (s.length() == 0) return;
    if (_entries.contains(s)) return; //If already in list then do nothing

    if (atBeginning) {
      _entries.add(0, s); //Place new entries at the front
      if (emitUpdate) emitIntervalAdded(0, 0); //All we did was add at loc 0
    }
    else {
      _entries.add(s); //Place new entries at end
      int loc=getSize() - 1;
      if (emitUpdate) emitIntervalAdded(loc, loc);
    }

    boolean trimmed=trimAsNeeded(); //Trim off excess items as required
    if (trimmed) if (emitUpdate) emitContentsChanged(0, getSize());

  }

  /** Add a listener to the list that's notified each time a change
   * to the data model occurs.
   * Note: Required for interface ListModel (part of ComboBoxModel)
   */
  public void addListDataListener(ListDataListener l) {
    if (_listeners == null) _listeners=new ArrayList();
    _listeners.add(l);
  }

  /**Just updates the whole list for simplicity (rather than sorting through all the individual changes*/
  public void emitContentsCompletelyChanged() {
    emitContentsChanged(0, getSize());
  }

  /**Emit changes to model to all ListDataListener objects.
   Note: Interval includes both indices*/
  private void emitContentsChanged(int ndx0, int ndx1) {
    if (_listeners == null) return;
    ListDataEvent lde=new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED,
        ndx0, ndx1);
    int len=_listeners.size();
    for (int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).contentsChanged(lde);
  }

  /**Emit changes to model to all ListDataListener objects
   Note: Interval includes both indices*/
  private void emitIntervalAdded(int ndx0, int ndx1) {
    if (_listeners == null) return;

    ListDataEvent lde;
    if (ndx0 == 0 && ndx1 == 0) lde=LDE_ADDED00; //usu. items are added at loc. 0
    else lde=new ListDataEvent(this, ListDataEvent.INTERVAL_ADDED, ndx0, ndx1);

    int len=_listeners.size();
    for (int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).intervalAdded(lde);
  }

  /**Emit changes to model to all ListDataListener objects
   Note: Interval includes both indices*/
  private void emitIntervalRemoved(int ndx0, int ndx1) {
    if (_listeners == null) return;
    ListDataEvent lde=new ListDataEvent(this, ListDataEvent.INTERVAL_REMOVED,
        ndx0, ndx1);
    int len=_listeners.size();
    for (int i=0; i < len; i++)
      ((ListDataListener)_listeners.get(i)).intervalRemoved(lde);
  }

  /**Returns the value at the specified index.
   Note: Required for interface ListModel (part of ComboBoxModel)*/
  public Object getElementAt(int ndx) {
    try { return _entries.get(ndx); }
    catch (Exception e) { return null; }
  }

  /**Return the selected item
   Note: Required for interface ComboBoxModel (part of MutableComboBoxModel)*/
  public Object getSelectedItem() {
    return _selectedItem;
  }

  /**Returns the length of the list.
   Note: Required for interface ListModel (part of ComboBoxModel)*/
  public int getSize() {
    return _entries.size();
  }

  /**Adds an item at a specific index
   Note: Required for interface MutableComboBoxModel*/
  public void insertElementAt(Object ob, int index) {
    addItem(ob, index, true);
  }

  /**Remove all items held in the combo box list
   Emits update to listeners
   */
  public void removeAllItems(boolean emitUpdate) {
    int len=getSize();
    if (len == 0) return; //Nothing to do
    for (int i=0; i < len; i++)
      _entries.remove(0);
    setSelectedItem(null);
    if (emitUpdate) emitIntervalRemoved(0, len); //Inform the listeners
  }

  /**Removes an item from the model
   Note: Required for interface MutableComboBoxModel*/
  public void removeElement(Object ob) {
    removeItem(ob, true);
  }

  /**Removes an item at a specific index
   Note: Required for interface MutableComboBoxModel*/
  public void removeElementAt(int index) {
    removeItem(index, true);
  }

  /**Remove specified item held in the combo box list
   Emits update to listeners if emitUpdate is true
   */
  public void removeItem(int which, boolean emitUpdate) {
    int len=getSize();
    if (which >= len) return; //Nothing to do
    _entries.remove(which);

    //JComboBox seems to require this:
    if (which == 0) setSelectedItem(getSize() <= 1 ? null
        : getElementAt(which + 1));
    else setSelectedItem(getElementAt(which - 1));

    if (emitUpdate) emitIntervalRemoved(which, which); //Inform the listeners
  }

  /**Remove specified item held in the combo box list
   Emits update to listeners if emitUpdate is true
   */
  public void removeItem(Object ob, boolean emitUpdate) {
    _entries.remove(ob);
    if (emitUpdate) emitContentsChanged(0, getSize()); //Inform the listeners
  }

  /**Remove a listener from the list that's notified each time a
   change to the data model occurs.
   Note: Required for interface ListModel (part of ComboBoxModel) */
  public void removeListDataListener(ListDataListener l) {
    if (_listeners == null) return;
    _listeners.remove(l);
  }

  /**Set the selected item
   Note: Required for interface ComboBoxModel (part of MutableComboBoxModel)*/
  public void setSelectedItem(Object anItem) {
    _selectedItem=anItem;
    addItem(_selectedItem, true, true);
    updatePersistentStore();
  }

  /**Trims the number of items held by this object
   (as specified by 'maxEntries')
   Returns true if any items were trimmed off
   */
  private boolean trimAsNeeded() {
    //See if need to trim the list:
    if (_maxEntries < 0) return false; //No trimming for max < 0
    int currentSize=getSize();
    boolean trimmed=false;
    while (currentSize > _maxEntries) {
      _entries.remove(currentSize - 1);
      trimmed=true;
      currentSize=getSize();
    }
    return trimmed;
  }

  /**Initializes this object from file or guru or ?*/
  private void initFromPersistentStore() {
    if (_name == null) return; //Needs a name in order for persistence to do its thing
    BufferedReader in=null;
    try {
      in=new BufferedReader(new FileReader(_name + _ext));
      while (in.ready()) {
        String s = in.readLine();
        if(okToIncludeItem(s))
          addItem(s, false, false);
      }
    }
    catch (FileNotFoundException e) {} //Ignore
    catch (Exception e) {
      System.out.println("Error reading in combo box persistence info for (" + _name + "): " + e);
    }
    finally { try { in.close(); } catch (Exception e) {} }
  }

  /**Updates persistent store with the current state of this object
   This version uses a file w/same name as object in the default directory
   */
  private void updatePersistentStore() {
    if (_name == null) return; //No name == no persistence
    BufferedWriter out=null;
    try {
      out=new BufferedWriter(new FileWriter(_name + _ext));
      int len=getSize();
      for (int i=0; i < len; i++) {
        Object ob=getElementAt(i);
        if (ob != null) {
          String s=ob.toString();
          if(okToIncludeItem(s)) {
            out.write(s);
            out.write("\n");
          }
        }
      }
    }
    catch (Exception e) {
      System.out.println("Error emitting combo box persistence info for ("+_name + "): " + e);
    }
    finally { try { out.close(); } catch (Exception e) {} }
  }

  /**Returns true if it's OK to include the specified item in the items
   * being written to or read from persistent store
   * Note: This gets overriden to change persistent behavior for this class*/
  protected boolean okToIncludeItem(String s) {
    if(s==null || s.length()==0) return false;
    return true;
  }
}
