package ri.util;

import java.io.File;
import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.List;


/** String Manipulation methods.  Primarily to ease Smalltalk to Java transition */
public class StringManip {

  private StringManip() {} //Private Constr: All methods in here are static

  /**Finds 'pattern' in 'data' beginning with the given 'start' index.
   * Knuth-Morris-Pratt Pattern Matching Algorithm. 
   * Returns location of match or -1
   * Note: This was adapted from function primMatchBytePattern (in RiApps7/ri.core.rtalk.RtPrimitives)
   * */
  public static int matchPattern(String data, String pattern, int start) {
    int[] failure=matchPatternFailCk(pattern);
    int j=0;
    int dlen=data.length();
    int plen=pattern.length();
    for(int i=start; i < dlen; i++) {
      while(j > 0 && pattern.charAt(j) != data.charAt(i))
        j=failure[j - 1];
      if(pattern.charAt(j) == data.charAt(i)) j++;
      if(j == plen) return i - plen + 1;
    }
    return -1;
  }

  /**Finds 'pattern' in 'data' beginning with the given 'start' index.
   * Knuth-Morris-Pratt Pattern Matching Algorithm. 
   * Returns location of match or -1
   * Note: This was adapted from function primMatchBytePattern (in RiApps7/ri.core.rtalk.RtPrimitives)
   * */
  public static int matchPattern(byte[] data, byte[] pattern, int start) {
    int[] failure=matchPatternFailCk(pattern);
    int j=0;
    int dlen=data.length;
    int plen=pattern.length;
    for(int i=start; i < dlen; i++) {
      while(j > 0 && pattern[j] != data[i])
        j=failure[j - 1];
      if(pattern[j] == data[i]) j++;
      if(j == plen) return i - plen + 1;
    }
    return -1;
  }

  /**Computes the failure function for Knuth-Morris-Pratt pattern match (used by matchPattern() above)*/
  private static int[] matchPatternFailCk(String pattern) {
    int plen=pattern.length();
    int[] failure=new int[plen];
    int j=0;
    for(int i=1; i < plen; i++) {
      while(j > 0 && pattern.charAt(j) != pattern.charAt(i))
        j=failure[j - 1];
      if(pattern.charAt(j) == pattern.charAt(i)) j++;
      failure[i]=j;
    }
    return failure;
  }

  /**Computes the failure function for Knuth-Morris-Pratt pattern match (used by matchPattern() above)*/
  private static int[] matchPatternFailCk(byte[] pattern) {
    int plen=pattern.length;
    int[] failure=new int[plen];
    int j=0;
    for(int i=1; i < plen; i++) {
      while(j > 0 && pattern[j] != pattern[i])
        j=failure[j - 1];
      if(pattern[j] == pattern[i]) j++;
      failure[i]=j;
    }
    return failure;
  }

  // -----------------------------------------------------------------------

  /**Returns given list as a csv (comma separated variable) string*/
  public static String getAsCsvString(List list) {
    int len=list == null ? 0 : list.size();
    if(len == 0) return ""; //no entries --> empty string
    Object ob=list.get(0);
    String s=(ob == null) ? "" : ob.toString();
    if(len == 1) return s; //only one entry, don't bother with a string buffer

    StringBuffer sb=new StringBuffer();
    sb.append(s);
    for(int i=1; i < len; i++) {
      sb.append(',');
      ob=list.get(i);
      s=(ob == null) ? "" : ob.toString();
      sb.append(s);
    }
    return sb.toString();
  }

  /**Returns a String consisting of the given number of spaces*/
  public static String spaces(int len) {
    if(len <= 0) return "";
    switch(len) {
      case 1: return " ";
      case 2: return "  ";
      case 3: return "   ";
      case 4: return "    ";
      case 5: return "     ";
      case 6: return "      ";
      case 7: return "       ";
      case 8: return "        ";
      case 9: return "         ";
      case 10:return "          ";
      default:
        StringBuffer sb=new StringBuffer(len);
        int len10 = len/10;
        for(int i=0; i < len10; i++)
          sb.append("          "); //do the big chunks
        sb.append(spaces(len - (len10 * 10))); //do the remainder
        return sb.toString();
    }
  }

  /**Returns a String consisting of the given number of zeroes*/
  public static String zeroes(int len) {
    if(len <= 0) return "";
    switch(len) {
      case 1: return "0";
      case 2: return "00";
      case 3: return "000";
      case 4: return "0000";
      case 5: return "00000";
      case 6: return "000000";
      case 7: return "0000000";
      case 8: return "00000000";
      case 9: return "000000000";
      case 10:return "0000000000";
      default:
        StringBuffer sb=new StringBuffer(len);
        int len10 = len/10;
        for(int i=0; i < len10; i++)
          sb.append("0000000000"); //do the big chunks
        sb.append(zeroes(len - (len10 * 10))); //do the remainder
        return sb.toString();
    }
  }

  /**Returns a string padded to the given width. 
   * If the string is longer than (or equal to) the specified width then the string itself is returned
   * null is ok.
   * */
  public static String padded(String s, int paddedLen) {
    int slen=s == null ? 0 : s.length();
    if(slen == 0) return spaces(paddedLen);
    if(slen >= paddedLen) return s;
    return s + spaces(paddedLen - slen);
  }

  /**Experimental thing: recursive toString() of array objects*/
  public static String arrayToString(Object arr) {
    if(arr == null || !arr.getClass().isArray()) return String.valueOf(arr);
    StringBuffer sb=new StringBuffer("[");
    int len=Array.getLength(arr);
    for(int i=0; i < len; i++) {
      if(i > 0) sb.append(',');
      sb.append(arrayToString(Array.get(arr, i)));
    }
    sb.append(']');
    return sb.toString();
  }

  /**Given a Carriage return delimited string.
     Returns as a list of strings (no carriage returns)*/
  public static List breakStringByCr(String s) {
    if(s == null) return null;
    ArrayList list=new ArrayList();
    int len=(s == null) ? 0 : s.length();
    int p1=0;
    int i;
    for(i=0; i < len; i++) {
      char ch=s.charAt(i);
      if(ch == '\n') {
        list.add(s.substring(p1, i));
        p1=i + 1;
        if(p1 < len && s.charAt(p1) == '\r') p1++;
      }
      else if(ch == '\r') ; //Ignore
    }
    if((i - p1) > 0) list.add(s.substring(p1, i)); //Last string
    return list;
  }

  /**Returns an ASCII viewable representation of the given byte array section.
   'buf' may be quite large but offset/len define the region to be printed.
   If len is -1 then its value is the full length of the buffer.
   If 'maxCnt' > 0 then is used as an upper limit on number of chars returned.
   Note: Non-ASCII bytes display as ~n where n is a hex value
   Note: Changed on 9/16/08: Used to display as decimal instead of hex
   */
  public static String clean(byte[] buf, int offset, int len, int maxCnt) {
    /* *** * /
    StringBuffer sb = new StringBuffer(new String(buf, 0, cnt));
    for(int i=0; i<cnt; i++)
      if(buf[i] < ' ' || buf[i] > '~') sb.setCharAt(i, '.');
    /* *** */
    if(buf == null) return "";
    int max=buf.length - offset;
    if(len < 0 || len > max) len=max;
    if(maxCnt > 0 && len > maxCnt) len=maxCnt; //Truncate
    StringBuffer sb=new StringBuffer();
    int stopAt=offset + len;
    for(int i=offset; i < stopAt; i++) {
      int b=buf[i];
      if(b < ' ' || b > '~') {
        sb.append("~");
        int val=b & 0xff;
        if(val <= 15) sb.append('0');
        sb.append(Integer.toHexString(val));
      }
      else sb.append((char)b);
    }
    return sb.toString();
  }

  /**Returns an ASCII viewable representation of the given byte array section.
   If 'maxCnt' > 0 then is used as an upper limit on number of chars returned.
   Note: Non-ASCII bytes display as periods '.'
   */
  public static String clean2(byte[] buf, int offset, int maxCnt) {
    if(buf == null) return "";
    int cnt=buf.length - offset;
    if(maxCnt > 0 && cnt > maxCnt) cnt=maxCnt; //Truncate
    StringBuffer sb=new StringBuffer(new String(buf, offset, cnt));
    cnt=sb.length(); //just in case
    for(int i=0; i < cnt; i++)
      if(buf[i] < ' ' || buf[i] > '~') sb.setCharAt(i, '.');
    return sb.toString();
  }

  /**Creates comma separated String from the given input String array.
     eg: "sa[0], sa[1] ..."*/
  public static String convertToString(String[] sa) {
    int cnt = sa==null ? 0 : sa.length;
    switch(cnt) {
      case 0: return ""; //no entries
      case 1: return sa[0]; //don't create StringBuffer for just one entry
      default:
        StringBuffer sb=new StringBuffer(sa[0]);
        for(int i=1; i < cnt; i++)
          sb.append(", ").append(sa[i]);
        return sb.toString();
    }
  }

  /**Returns just the base name in a pathname**/
  public static String extractBasename(String s) {
    if(s == null) return null;
    int pos=s.lastIndexOf(File.separatorChar);
    if(pos == s.length() - 1) return ""; //no basename
    if(pos >= 0) return s.substring(pos + 1);
    return s; //return whole thing if no path found
  }

  /**Returns just the path portion (ie: "C:\pathanme\")*/
  public static String extractPath(String s) {
    if(s == null) return null;
    char sepChar=File.separatorChar;
    int pos=s.lastIndexOf(sepChar);
    if(pos >= 0) return s.substring(0, pos) + sepChar;
    return "";
  }

  /**Returns specified field in the String (1 based)*/
  public static String getField(int which, String s) {
    int pos1=0;
    int pos2=-1;
    ///System.out.println("s="+s); // ***
    try {
      for(int i=0; i < which; i++) {
        pos2=s.indexOf(',', pos1);
        if(i == (which - 1)) return s.substring(pos1, pos2);
        pos1=pos2 + 1;
      }
    }
    catch(StringIndexOutOfBoundsException e) {}
    catch(NullPointerException e) {}
    return "";
  }

  /**Returns the index of the first String in 'sa' that matches 's'
   *      If ignoreCase==true then compare is case insensitive.
   *      Returns -1 on no match.
   */
  public static int getIndexOf(String s, String[] sa, boolean ignoreCase) {
    if(sa == null) return -1;
    if(s == null) return -1;
    int i;
    int length=sa.length;
    if(ignoreCase) {
      for(i=0; i < length; i++)
        if(sa[i].equalsIgnoreCase(s)) return i;
    }
    else for(i=0; i < length; i++)
      if(sa[i].equals(s)) return i;
    return -1;
  }

  /**Returns string with all blanks removed*/
  public static String removeBlanks(String s) {
    if(s == null || s.length() == 0) return s;
    int pos=s.indexOf(' '); //SPACE
    if(pos < 0) pos=s.indexOf(9); //TAB
    if(pos < 0) return s;
    StringBuffer sb=new StringBuffer();
    int length=s.length();
    for(int i=0; i < length; i++) {
      char ch=s.charAt(i);
      if(!Character.isSpaceChar(ch)) sb.append(ch);
    }
    return sb.toString();
  }

  /**Converts an array of bytes into a nice comma separated string of (decimal) numbers*/
  public static String stringIt(byte[] ba) {
    if(ba == null || ba.length == 0) return "";
    StringBuffer sb=new StringBuffer();
    int length=ba.length;
    for(int i=0; i < length; i++) {
      sb.append("" + (ba[i] & 0xff));
      if(i != length - 1) sb.append(",");
    }
    return sb.toString();
  }

  /**Strip off the pathname and extension returning only the base filename*/
  public static String stripPathExt(String s) {
    if(s == null) return null;
    int pos=s.lastIndexOf('\\');
    if(pos < 0) pos=s.lastIndexOf('/');
    if(pos >= 0) s=s.substring(pos + 1); //Path removed
    pos=s.lastIndexOf('.');
    if(pos >= 0) return s.substring(0, pos); //extension removed
    return s;
  }

  /**Trims leading and trailing spaces from a string.
     Note: This really only does spaces (not whitespace).
           use String.trim() to trim whitespace as well as
           control characters.*/
  public static String trim(String s) {
    if(s == null) return null;

    int len=s.length();
    int pos1=0;
    for(; pos1 < len && s.charAt(pos1) == ' '; pos1++); //Find leading spaces
    if(pos1 >= len) return ""; //All spaces
    int pos2=len;
    for(; pos2 > 0 && s.charAt(pos2 - 1) == ' '; pos2--); //Find trailing spaces

    if(pos1 == 0 && pos2 == len) return s; //Nothing done
    return s.substring(pos1, pos2); //Spaces removed
  }

  /**Trims off leading whitespace*/
  public static String trimBlanks(String s) {
    if(s == null || s.length() < 1) return s;
    if(Character.isSpaceChar(s.charAt(0))) { //If starts with a space
      int i;
      int len=s.length();
      for(i=1; (i < len) && (Character.isSpaceChar(s.charAt(i))); i++); //step past leading blanks
      return s.substring(i);
    }
    return s;
  }

  /**Converts the given String to an array of Strings using the given
     character as the separation delimiter*/
  public static String[] toArray(String s, char ch) {
    if(s == null || s.length() < 1) return null;

    //Determine number of fields that will be needed:
    int sepCnt=0;
    int length=s.length();
    for(int i=0; i < length; i++)
      if(s.charAt(i) == ch) sepCnt++;
    String[] sa=new String[sepCnt + 1];

    //Load 'em up:
    int cnt=0;
    int curPos;
    int newPos=-1;
    ///int endPos = s.length()-1;
    do {
      curPos=newPos + 1;
      newPos=s.indexOf(ch, curPos);
      if(newPos >= 0) sa[cnt++]=s.substring(curPos, newPos);
      else sa[cnt++]=s.substring(curPos); //last item
    }
    while(newPos >= 0);
    return sa;
  }

  /**Adds the given prefixes to the given CSV String
     Purpose is to create Guru keys out of simple names
     Input is a String array where the first index is the
     'Names' sequence and the second index is the 'prefixes'
     For example {"Name,Val"} {"ri.sys.", "hw."}
     becomes "ri.sys.Name", "hw.val"

     @seeAlso removePrefixesFromCsv
   */

  public static String addPrefixesToCsv(String[] combined) {
    int clen=(combined == null) ? 0 : combined.length;
    if(clen == 0) return null;
    if(clen == 1) return combined[0]; //Nothing to do

    String[] names=Csv.toArray(combined[0]);
    String[] prefixes=Csv.toArray(combined[1]);

    int nLen=names.length;
    int pLen=prefixes.length;
    int len=Math.min(nLen, pLen);

    StringBuffer sb=new StringBuffer();

    if(len > 0) sb.append(prefixes[0]).append(names[0]);
    for(int i=1; i < len; i++)
      sb.append(',').append(prefixes[i]).append(names[i]);

    return sb.toString();
  }

  /**Adds in the given 'setTo' string to the
     beginnings of each CSV entry in the given source string
     Note: The 'src' string is taken as a comma separated (CSV) set of values
     Example: addToCsv("Name,Val", "ri.sys.")
     will return: ri.sys.Name,ri.sys.Val"

     @seeAlso removeFromCsv
   */

  public static String addToCsv(String src, String setPrefixTo) {
    if(setPrefixTo == null) return src;
    if(src == null) return src;

    String[] sa=Csv.toArray(src);
    int len=(sa == null) ? 0 : sa.length;
    if(len == 0) return "";

    for(int i=0; i < len; i++)
      sa[i]=setPrefixTo + sa[i];
    return Csv.toCsv(sa);
  }

  /**Removes instances of the given 'searchFor' string from the
     beginnings of each CSV entry in the given source string
     Note: The 'src' string is taken as a comma separated (CSV) set of values
     Example: removeFromCsv("ri.sys.Name,ri.sys.Val", "ri.sys.")
     will return: "Name,Val"

     @seeAlso addToCsv
   */

  public static String removeFromCsv(String src, String searchFor) {
    if(searchFor == null) return src;
    if(src == null) return src;

    String[] sa=Csv.toArray(src);
    int len=(sa == null) ? 0 : sa.length;
    if(len == 0) return "";

    int startPos=searchFor.length();
    for(int i=0; i < len; i++) {
      String s=sa[i];
      if(s.startsWith(searchFor)) sa[i]=s.substring(startPos);
    }
    return Csv.toCsv(sa);
  }

  /**Removes all prefixes from the given Guru Keys
     For example "ri.sys.Name,hw.Val"
     becomes to strings: "Name,Val" and "ri.sys.,hw."
     Returns a two element String array
     The first index is the cleaned up names
     The second is the prefixes that were removed

     @seeAlso addPrefixesToCsv
   */

  public static String[] removePrefixesFromCsv(String src) {
    String[] sa=Csv.toArray(src);
    int len=(sa == null) ? 0 : sa.length;

    StringBuffer names=new StringBuffer();
    StringBuffer prefixes=new StringBuffer();

    for(int i=0; i < len; i++) {
      String s=sa[i];
      int pos=s.lastIndexOf('.');
      if(pos >= 0) {
        prefixes.append(s.substring(0, pos + 1));
        if(s.length() >= (pos + 1)) names.append(s.substring(pos + 1));
      }
      else names.append(s);

      if(i < len - 1) {
        prefixes.append(',');
        names.append(',');
      }
    }

    return new String[] { names.toString(), prefixes.toString() };
  }

}
